from __future__ import annotations

import procrunner

from dials.array_family import flex


def test_import_integrate_hkl(dials_data, tmp_path):
    data = dials_data("centroid_test_data", pathlib=True)
    result = procrunner.run(
        [
            "dials.import_xds",
            "input.method=reflections",
            data / "INTEGRATE.HKL",
            data / "experiments.json",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    table = flex.reflection_table.from_file(tmp_path / "integrate_hkl.refl")
    assert "miller_index" in table
    assert "id" in table
    assert "panel" in table
    assert "xyzcal.px" in table
    assert "xyzobs.px.value" in table
    assert "intensity.cor.value" in table
    assert "intensity.cor.variance" in table
    assert len(table) == 174911


def test_import_spot_xds(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.import_xds",
            "input.method=reflections",
            dials_data("centroid_test_data", pathlib=True) / "SPOT.XDS",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    table = flex.reflection_table.from_file(tmp_path / "spot_xds.refl")
    assert "miller_index" in table
    assert "id" in table
    assert "panel" in table
    assert "xyzobs.px.value" in table
    assert "intensity.sum.value" in table
    assert len(table) == 742


def test_import_spot_xds_with_filtering(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.import_xds",
            "input.method=reflections",
            dials_data("centroid_test_data", pathlib=True) / "SPOT.XDS",
            "remove_invalid=True",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    table = flex.reflection_table.from_file(tmp_path / "spot_xds.refl")
    assert "miller_index" in table
    assert "id" in table
    assert "panel" in table
    assert "xyzobs.px.value" in table
    assert "intensity.sum.value" in table
    assert len(table) == 664


def test_from_xds_files(dials_data, tmp_path):
    # Import from the image files
    result = procrunner.run(
        [
            "dials.import_xds",
            "input.method=experiment",
            "output.filename=import_xds.expt",
            dials_data("centroid_test_data", pathlib=True),
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert tmp_path.joinpath("import_xds.expt").is_file()



from __future__ import annotations

import math
import os

import pytest

from scitbx import matrix


class SpotPredictor:
    def __init__(self, dials_regression):
        import dxtbx
        from iotbx.xds import integrate_hkl, xparm
        from rstbx.cftbx.coordinate_frame_converter import coordinate_frame_converter

        from dials.algorithms.spot_prediction import (
            IndexGenerator,
            ScanStaticRayPredictor,
            ray_intersection,
        )
        from dials.util import ioutil

        # The XDS files to read from
        integrate_filename = os.path.join(
            dials_regression, "data", "sim_mx", "INTEGRATE.HKL"
        )
        gxparm_filename = os.path.join(dials_regression, "data", "sim_mx", "GXPARM.XDS")

        # Read the XDS files
        self.integrate_handle = integrate_hkl.reader()
        self.integrate_handle.read_file(integrate_filename)
        self.gxparm_handle = xparm.reader()
        self.gxparm_handle.read_file(gxparm_filename)

        # Get the parameters we need from the GXPARM file
        models = dxtbx.load(gxparm_filename)
        self.beam = models.get_beam()
        self.gonio = models.get_goniometer()
        self.detector = models.get_detector()
        self.scan = models.get_scan()

        assert len(self.detector) == 1

        # print self.detector

        # Get crystal parameters
        self.space_group_type = ioutil.get_space_group_type_from_xparm(
            self.gxparm_handle
        )
        cfc = coordinate_frame_converter(gxparm_filename)
        a_vec = cfc.get("real_space_a")
        b_vec = cfc.get("real_space_b")
        c_vec = cfc.get("real_space_c")
        self.unit_cell = cfc.get_unit_cell()
        self.ub_matrix = matrix.sqr(a_vec + b_vec + c_vec).inverse()

        # Get the minimum resolution in the integrate file
        self.d_min = self.detector[0].get_max_resolution_at_corners(self.beam.get_s0())

        # Get the number of frames from the max z value
        xcal, ycal, zcal = zip(*self.integrate_handle.xyzcal)
        self.scan.set_image_range(
            (
                self.scan.get_image_range()[0],
                self.scan.get_image_range()[0] + int(math.ceil(max(zcal))),
            )
        )

        # Create the index generator
        generate_indices = IndexGenerator(
            self.unit_cell, self.space_group_type, self.d_min
        )

        s0 = self.beam.get_s0()
        m2 = self.gonio.get_rotation_axis()
        fixed_rotation = self.gonio.get_fixed_rotation()
        setting_rotation = self.gonio.get_setting_rotation()
        UB = self.ub_matrix
        dphi = self.scan.get_oscillation_range(deg=False)

        # Create the ray predictor
        self.predict_rays = ScanStaticRayPredictor(
            s0, m2, fixed_rotation, setting_rotation, dphi
        )

        # Predict the spot locations
        self.reflections = self.predict_rays(generate_indices.to_array(), UB)

        # Calculate the intersection of the detector and reflection frames
        success = ray_intersection(self.detector, self.reflections)
        self.reflections.select(success)


@pytest.fixture(scope="session")
def spotpredictor(dials_regression):
    return SpotPredictor(dials_regression)


def test_dmin(spotpredictor):
    """Ensure calculated d_min < d_min in integrate file"""
    d = [spotpredictor.unit_cell.d(h) for h in spotpredictor.integrate_handle.hkl]
    d_min = min(d)
    assert spotpredictor.d_min <= d_min


def test_miller_index_set(spotpredictor):
    """Ensure we have the whole set of miller indices"""
    gen_hkl = {}
    for r in spotpredictor.reflections.rows():
        gen_hkl[r["miller_index"]] = True
    for hkl in spotpredictor.integrate_handle.hkl:
        assert gen_hkl[hkl]


def test_rotation_angles(spotpredictor):
    """Ensure the rotation angles agree with XDS"""

    # Create a dict of lists of xy for each hkl
    gen_phi = {}
    for r in spotpredictor.reflections.rows():
        hkl = r["miller_index"]
        phi = r["phi"]
        try:
            a = gen_phi[hkl]
            a.append(phi)
            gen_phi[hkl] = a
        except KeyError:
            gen_phi[hkl] = [phi]

    for hkl, xyz in zip(
        spotpredictor.integrate_handle.hkl, spotpredictor.integrate_handle.xyzcal
    ):

        xds_phi = (
            spotpredictor.scan.get_oscillation(deg=False)[0]
            + xyz[2] * spotpredictor.scan.get_oscillation(deg=False)[1]
        )

        # Select the nearest xy to use if there are 2
        my_phi = gen_phi[hkl]
        if len(my_phi) == 2:
            my_phi0 = my_phi[0]
            my_phi1 = my_phi[1]
            diff0 = abs(xds_phi - my_phi0)
            diff1 = abs(xds_phi - my_phi1)
            if diff0 < diff1:
                my_phi = my_phi0
            else:
                my_phi = my_phi1
        else:
            my_phi = my_phi[0]

        assert xds_phi == pytest.approx(my_phi, abs=0.1)


def test_beam_vectors(spotpredictor):
    """Ensure |s1| == |s0|"""
    s0_length = matrix.col(spotpredictor.beam.get_s0()).length()
    for r in spotpredictor.reflections.rows():
        s1 = r["s1"]
        s1_length = matrix.col(s1).length()
        assert s0_length == pytest.approx(s1_length, abs=1e-7)


def test_image_coordinates(spotpredictor):
    """Ensure the image coordinates agree with XDS"""
    # Create a dict of lists of xy for each hkl
    gen_xy = {}
    for r in spotpredictor.reflections.rows():
        hkl = r["miller_index"]
        xy = r["xyzcal.mm"][0:2]
        xy = spotpredictor.detector[0].millimeter_to_pixel(xy)
        try:
            a = gen_xy[hkl]
            a.append(xy)
            gen_xy[hkl] = a
        except KeyError:
            gen_xy[hkl] = [xy]

    for hkl, xyz in zip(
        spotpredictor.integrate_handle.hkl, spotpredictor.integrate_handle.xyzcal
    ):

        xds_xy = (xyz[0] - 0.5, xyz[1] - 0.5)

        # Select the nearest xy to use if there are 2
        my_xy = gen_xy[hkl]
        if len(my_xy) == 2:
            my_xy0 = my_xy[0]
            my_xy1 = my_xy[1]
            diff0 = (matrix.col(xds_xy) - matrix.col(my_xy0)).length()
            diff1 = (matrix.col(xds_xy) - matrix.col(my_xy1)).length()
            if diff0 < diff1:
                my_xy = my_xy0
            else:
                my_xy = my_xy1
        else:
            my_xy = my_xy[0]

        assert xds_xy[0] == pytest.approx(my_xy[0], abs=0.1), (xds_xy, gen_xy[hkl])
        assert xds_xy[1] == pytest.approx(my_xy[1], abs=0.1), (xds_xy, gen_xy[hkl])



from __future__ import annotations

import random

from cctbx import sgtbx
from scitbx.array_family import flex

from dials.algorithms.clustering import plots
from dials.algorithms.clustering.unit_cell import cluster_unit_cells


def test_plot_uc_histograms():
    sgi = sgtbx.space_group_info("P1")
    unit_cells = [
        sgi.any_compatible_unit_cell(volume=random.uniform(990, 1010)).parameters()
        for i in range(10)
    ]
    uc_params = [flex.double(a) for a in zip(*unit_cells)]
    d = plots.plot_uc_histograms(uc_params)
    assert set(d) == {"uc_scatter", "uc_hist"}
    assert len(d["uc_scatter"]["data"]) == 3
    assert len(d["uc_hist"]["data"]) == 3


def test_scipy_dendrogram_to_plotly_json():
    # generate some random unit cells
    sgi = sgtbx.space_group_info("P1")
    crystal_symmetries = [
        sgi.any_compatible_crystal_symmetry(volume=random.uniform(990, 1010))
        for i in range(10)
    ]
    lattice_ids = list(range(len(crystal_symmetries)))
    result = cluster_unit_cells(crystal_symmetries, lattice_ids)
    d = plots.scipy_dendrogram_to_plotly_json(
        result.dendrogram, title="Unit cell clustering"
    )
    assert set(d) == {"layout", "data"}



"""
Functions to help with calculating batch properties for experiments objects.
"""

from __future__ import annotations

import logging

from dials.array_family import flex

logger = logging.getLogger("dials")


class batch_manager:
    def __init__(self, batches, batch_params):
        # batch params is a list of dicts with "id" and "range" - used to be
        # a 'scope extract' object
        self.batch_params = sorted(batch_params, key=lambda b: b["range"][0])
        self.batches = batches
        self.reduced_batches, self._batch_increments = self._reduce()

    def _reduce(self):
        reduced_batches = flex.int(self.batches)
        batch_increments = []
        incr = 0
        for batch in self.batch_params:
            sel = (reduced_batches >= batch["range"][0]) & (
                reduced_batches <= batch["range"][1]
            )
            reduced_batches.set_selected(
                sel, reduced_batches.select(sel) - (batch["range"][0] - incr) + 1
            )
            batch_increments.append(incr)
            incr += batch["range"][1] - batch["range"][0] + 1
        assert len(set(reduced_batches)) == len(reduced_batches)
        return list(reduced_batches), batch_increments

    def batch_plot_shapes_and_annotations(self):
        light_grey = "#d3d3d3"
        grey = "#808080"
        shapes = []
        annotations = []
        batches = flex.int(self.batches)
        text = flex.std_string(batches.size())
        for i, batch in enumerate(self.batch_params):
            fillcolor = [light_grey, grey][i % 2]  # alternate colours
            shapes.append(
                {
                    "type": "rect",
                    # x-reference is assigned to the x-values
                    "xref": "x",
                    # y-reference is assigned to the plot paper [0,1]
                    "yref": "paper",
                    "x0": self._batch_increments[i],
                    "y0": 0,
                    "x1": self._batch_increments[i]
                    + (batch["range"][1] - batch["range"][0]),
                    "y1": 1,
                    "fillcolor": fillcolor,
                    "opacity": 0.2,
                    "line": {"width": 0},
                }
            )
            annotations.append(
                {
                    # x-reference is assigned to the x-values
                    "xref": "x",
                    # y-reference is assigned to the plot paper [0,1]
                    "yref": "paper",
                    "x": self._batch_increments[i]
                    + (batch["range"][1] - batch["range"][0]) / 2,
                    "y": 1,
                    "text": f"{batch['id']}",
                    "showarrow": False,
                    "yshift": 20,
                    # 'arrowhead': 7,
                    # 'ax': 0,
                    # 'ay': -40
                }
            )
            sel = (batches >= batch["range"][0]) & (batches <= batch["range"][1])
            text.set_selected(
                sel,
                flex.std_string(
                    [
                        f"{batch['id']}: {j - batch['range'][0] + 1}"
                        for j in batches.select(sel)
                    ]
                ),
            )
        return shapes, annotations, list(text)


def assign_batches_to_reflections(reflections, batch_offsets):
    """Assign a 'batch' column to the reflection table"""
    for batch_offset, refl in zip(batch_offsets, reflections):
        xdet, ydet, zdet = [flex.double(x) for x in refl["xyzobs.px.value"].parts()]
        # compute BATCH values - floor() to get (fortran) image captured within
        #                        +1     because FORTRAN counting; zdet+1=image_index
        #                        +off   because            image_index+o=batch
        refl["batch"] = (flex.floor(zdet).iround() + 1) + batch_offset
    return reflections


def get_batch_ranges(experiments, batch_offsets):
    """Get batch ranges for a list of experiments and offsets"""
    batch_ranges = []
    assert len(experiments) == len(batch_offsets)
    image_ranges = get_image_ranges(experiments)
    for batch_offset, image_range in zip(batch_offsets, image_ranges):
        batch_ranges.append(
            (image_range[0] + batch_offset, image_range[1] + batch_offset)
        )
    return batch_ranges


def get_image_ranges(experiments):
    """Get image ranges for a list of experiments (including scanless exp.)"""
    # Note, if set to 1,1,for scanless experiments then first batch offset in
    # _calculate_batch_offsets is zero below, bad!
    return [e.scan.get_image_range() if e.scan else (0, 0) for e in experiments]


def calculate_batch_offsets(experiment_list):
    """Take a list of experiments and resolve and return the batch offsets.
    First adds an image_range property as not all experiments have scans."""
    image_ranges = get_image_ranges(experiment_list)
    offsets = _calculate_batch_offsets(image_ranges)
    return offsets


def set_batch_offsets(experiment_list, batch_offsets):
    """Set batch offsets in scan objects. Don't need to set anything for
    scanless experiments, as these are not used with the batch system."""
    for exp, offset in zip(experiment_list, batch_offsets):
        if exp.scan:
            exp.scan.set_batch_offset(offset)


def _calculate_batch_offsets(image_ranges):
    """Take a list of (modified) experiments and resolve and return the batch
    offsets.

    This is the number added to the image number to give the
    batch number, such that:
    - Each experiment has a unique, nonoverlapping, nonconsecutive range
    - None are zero
    - Image number ranges are kept if at all possible
    """

    experiments_to_shift = []
    existing_ranges = set()
    maximum_batch_number = 0
    batch_offsets = [0] * len(image_ranges)

    # Handle zeroth shifts and kept ranges
    for i, image_range in enumerate(image_ranges):
        ilow, ihigh = image_range
        # Check assumptions
        assert ilow <= ihigh, "Inverted image order!?"
        assert ilow >= 0, "Negative image indices are not expected"
        # Don't emit zero: Causes problems with C/fortran number conversion
        if ilow == 0:
            ilow, ihigh = ilow + 1, ihigh + 1
        # If we overlap with anything, then process later
        if any(ilow < high + 1 and ihigh >= low - 1 for low, high in existing_ranges):
            experiments_to_shift.append((i, image_range))
        else:
            batch_offsets[i] = ilow - image_range[0]
            existing_ranges.add((ilow, ihigh))
            maximum_batch_number = max(maximum_batch_number, ihigh)
    # Now handle all the experiments that overlapped by pushing them higher
    for i, image_range in experiments_to_shift:
        start_number = _next_epoch(maximum_batch_number)
        range_width = image_range[1] - image_range[0] + 1
        end_number = start_number + range_width - 1
        batch_offsets[i] = start_number - image_range[0]
        maximum_batch_number = end_number
    return batch_offsets


def _next_epoch(val):
    """Find the next number above the existing value that ends in 1, that is
    not consecutive with the current value."""
    if val % 100 == 99:
        return val + 2
    elif val % 100 == 0:
        return val + 101
    else:
        rem = val % 100
        return val - rem + 101



from __future__ import annotations

import math


class NeighborAnalysis:
    def __init__(
        self,
        reflections,
        step_size=45,
        tolerance=1.5,
        max_height_fraction=0.25,
        percentile=None,
        histogram_binning="linear",
        nn_per_bin=5,
    ):
        self.tolerance = tolerance  # Margin of error for max unit cell estimate
        from scitbx.array_family import flex

        NEAR = 10
        self.NNBIN = nn_per_bin  # target number of neighbors per histogram bin
        self.histogram_binning = histogram_binning

        direct = flex.double()

        if "entering" in reflections:
            entering_flags = reflections["entering"]
        else:
            entering_flags = flex.bool(reflections.size(), True)
        rs_vectors = reflections["rlp"]
        phi_deg = reflections["xyzobs.mm.value"].parts()[2] * (180 / math.pi)

        d_spacings = flex.double()
        # nearest neighbor analysis
        from annlib_ext import AnnAdaptor

        for imageset_id in range(flex.max(reflections["imageset_id"]) + 1):
            sel_imageset = reflections["imageset_id"] == imageset_id
            if sel_imageset.count(True) == 0:
                continue
            phi_min = flex.min(phi_deg.select(sel_imageset))
            phi_max = flex.max(phi_deg.select(sel_imageset))
            d_phi = phi_max - phi_min
            n_steps = max(int(math.ceil(d_phi / step_size)), 1)

            for n in range(n_steps):
                sel_step = (
                    sel_imageset
                    & (phi_deg >= (phi_min + n * step_size))
                    & (phi_deg < (phi_min + (n + 1) * step_size))
                )

                for entering in (True, False):
                    sel_entering = sel_step & (entering_flags == entering)
                    if sel_entering.count(True) == 0:
                        continue

                    query = flex.double()
                    query.extend(rs_vectors.select(sel_entering).as_double())

                    if query.size() == 0:
                        continue

                    IS_adapt = AnnAdaptor(data=query, dim=3, k=1)
                    IS_adapt.query(query)

                    direct.extend(1 / flex.sqrt(IS_adapt.distances))
                    d_spacings.extend(1 / rs_vectors.norms())

        assert (
            len(direct) > NEAR
        ), f"Too few spots ({len(direct)}) for nearest neighbour analysis."

        perm = flex.sort_permutation(direct)
        direct = direct.select(perm)
        d_spacings = d_spacings.select(perm)

        # eliminate nonsensical direct space distances
        sel = direct > 1
        direct = direct.select(sel)
        d_spacings = d_spacings.select(sel)

        if percentile is None:
            # reject top 1% of longest distances to hopefully get rid of any outliers
            n = int(math.floor(0.99 * len(direct)))
            direct = direct[:n]
            d_spacings = d_spacings[:n]

        # determine the most probable nearest neighbor distance (direct space)
        if self.histogram_binning == "log":
            hst = flex.histogram(
                flex.log10(direct), n_slots=int(len(direct) / self.NNBIN)
            )
        else:
            hst = flex.histogram(direct, n_slots=int(len(direct) / self.NNBIN))
        if self.histogram_binning == "log":
            self.slot_start = flex.double(
                [10 ** (s - 0.5 * hst.slot_width()) for s in hst.slot_centers()]
            )
            self.slot_end = flex.double(
                [10 ** (s + 0.5 * hst.slot_width()) for s in hst.slot_centers()]
            )
            self.slot_width = self.slot_end - self.slot_start
        else:
            self.slot_start = hst.slot_centers() - 0.5 * hst.slot_width()
            self.slot_end = hst.slot_centers() + 0.5 * hst.slot_width()
            self.slot_width = hst.slot_width()
        self.relative_frequency = hst.slots().as_double() / self.slot_width
        highest_bin_height = flex.max(self.relative_frequency)

        if percentile is not None:
            # determine the nth-percentile direct-space distance
            perm = flex.sort_permutation(direct, reverse=True)
            self.max_cell = (
                self.tolerance * direct[perm[int((1 - percentile) * len(direct))]]
            )

        else:
            # choose a max cell based on bins above a given fraction of the highest bin height
            # given multiple
            isel = (
                self.relative_frequency.as_double()
                > (max_height_fraction * highest_bin_height)
            ).iselection()
            self.max_cell = (
                self.tolerance * self.slot_end[int(flex.max(isel.as_double()))]
            )

        self.reciprocal_lattice_vectors = rs_vectors
        self.d_spacings = d_spacings
        self.direct = direct
        self.histogram = hst

    def plot_histogram(self, filename="nn_hist.png", figsize=(12, 8)):
        import matplotlib.pyplot as plt

        plt.figure(figsize=figsize)
        plt.bar(
            self.slot_start,
            self.relative_frequency,
            align="center",
            width=self.slot_width,
            color="black",
            edgecolor=None,
        )
        ymin, ymax = plt.ylim()
        if self.histogram_binning == "log":
            ax = plt.gca()
            ax.set_xscale("log")
        plt.vlines(
            self.max_cell / self.tolerance,
            ymin,
            ymax,
            linestyles="--",
            colors="g",
            label="estimated max cell",
        )
        plt.vlines(
            self.max_cell,
            ymin,
            ymax,
            colors="g",
            label="estimated max cell (including tolerance)",
        )
        plt.xlabel("Direct space distance (A)")
        plt.ylabel("Frequency")
        plt.legend(loc="upper left")
        plt.savefig(filename)
        plt.clf()



from __future__ import annotations

import json

import procrunner
import pytest

from dials.array_family import flex
from dials.command_line import detect_blanks


def test_strong(dials_data, capsys, run_in_tmp_path):
    expts_file = dials_data("insulin_processed", pathlib=True) / "imported.expt"
    refl_file = dials_data("insulin_processed", pathlib=True) / "strong.refl"
    refl = flex.reflection_table.from_file(refl_file)
    z = refl["xyzobs.px.value"].parts()[2]
    refl_subset = refl.select((z < 10) | (z > 20))
    refl_subset.as_file(run_in_tmp_path / "mod.refl")

    detect_blanks.run([str(expts_file), str(run_in_tmp_path / "mod.refl")])

    captured = capsys.readouterr()
    assert "Potential blank images: 11 -> 21" in captured.out
    json_file = run_in_tmp_path / "blanks.json"
    assert json_file.is_file()
    with json_file.open() as fh:
        d = json.load(fh)
    assert d["strong"]["blank_regions"] == [[10, 21]]
    assert not d["indexed"]
    assert not d["integrated"]


def test_integrated(dials_data, capsys, run_in_tmp_path):
    expts_file = dials_data("insulin_processed", pathlib=True) / "integrated.expt"
    refl_file = dials_data("insulin_processed", pathlib=True) / "integrated.refl"
    refl = flex.reflection_table.from_file(refl_file)
    z = refl["xyzobs.px.value"].parts()[2]
    sel = z < 40
    refl_subset = refl.select(sel)
    refl_subset["intensity.prf.value"].set_selected(
        z.select(sel) < 10, refl_subset["intensity.prf.value"] * 0.05
    )
    refl_subset.as_file(run_in_tmp_path / "mod.refl")

    detect_blanks.run(
        [
            str(expts_file),
            str(run_in_tmp_path / "mod.refl"),
            "output.json=detect_blanks.json",
        ]
    )

    captured = capsys.readouterr()
    assert (
        """\
Analysis of 18171 strong reflections:
Potential blank images: 41 -> 45
Analysis of 18171 indexed reflections:
Potential blank images: 41 -> 45
Analysis of 44340 integrated reflections:
Potential blank images: 1 -> 9
Potential blank images: 41 -> 45
"""
        in captured.out
    )
    json_file = run_in_tmp_path / "detect_blanks.json"
    assert json_file.is_file()
    with json_file.open() as fh:
        d = json.load(fh)
    assert d["strong"]["blank_regions"] == [[40, 45]]
    assert d["indexed"]["blank_regions"] == [[40, 45]]
    assert d["integrated"]["blank_regions"] == [[0, 9], [40, 45]]


def test_passing_still_images_raises_sysexit(dials_data, run_in_tmp_path):
    path = dials_data("thaumatin_grid_scan", pathlib=True)

    # Import the data
    result = procrunner.run(
        ["dials.import", "output.experiments=imported.expt"]
        + list(path.glob("*.cbf*")),
        working_directory=run_in_tmp_path,
    )
    assert not result.returncode and not result.stderr
    expts_file = run_in_tmp_path / "imported.expt"
    assert expts_file.is_file()

    # Find the spots
    result = procrunner.run(
        ["dials.find_spots", "imported.expt", "min_spot_size=3", "nproc=1"],
        working_directory=run_in_tmp_path,
    )
    assert not result.returncode and not result.stderr
    refl_file = run_in_tmp_path / "strong.refl"
    assert refl_file.is_file()

    # Now test that passing stills raises a sys.exit
    with pytest.raises(SystemExit):
        detect_blanks.run([str(expts_file), str(refl_file)])



"""Calculations relevant to reporting."""

from __future__ import annotations

from cctbx import miller
from scitbx.array_family import flex

from dials.algorithms.scaling.scaling_library import scaled_data_as_miller_array
from dials.util.batch_handling import (
    assign_batches_to_reflections,
    calculate_batch_offsets,
    get_batch_ranges,
)


def batch_dependent_properties(batches, intensities, scales=None):
    """
    Calculate intensity properties as a function of batch.

    Manage individual calculations to ensure that all calculations are
    done on the same batches array and consistent length lists returned.

    Args:
        batches (miller array): the batch numbers for the reflections.
        intensities (miller array): the reflection intensities, with
            sigmas also present.
        scales (miller array, optional): the scale factors of the reflections.

    Returns:
        (tuple): tuple containing:
            binned_batches (list): list of batch number bins
            rmerge (list): rmerge for each bin
            isigi (list): average I/sigma for each bin
            scales (list): average scale for each bin if scales given, else None

    Raises:
        AssertionError: Raised if sigmas not present in intensities, or if
            arrays sizes do not match.
    """
    assert intensities.size() == batches.size()
    if scales:
        assert scales.size() == batches.size()
    assert intensities.sigmas() is not None
    sel = intensities.sigmas() > 0
    intensities = intensities.select(sel)
    batches = batches.select(sel)
    if scales:
        scales = scales.select(sel)

    binned_batches, rmerge = rmerge_vs_batch(intensities, batches)
    _, isigi = i_sig_i_vs_batch(intensities, batches)

    scalesvsbatch = None
    if scales:
        _, scalesvsbatch = scales_vs_batch(scales, batches)
    return binned_batches, rmerge, isigi, scalesvsbatch


def combined_table_to_batch_dependent_properties(
    combined_table, experiments, scaled_array=None
):
    """Extract batch dependent properties from a combined reflection table."""
    tables = []
    for id_ in set(combined_table["id"]).difference({-1}):
        tables.append(combined_table.select(combined_table["id"] == id_))

    return reflection_tables_to_batch_dependent_properties(
        tables, experiments, scaled_array
    )


def reflection_tables_to_batch_dependent_properties(
    reflection_tables, experiments, scaled_array=None
):
    """Extract batch dependent properties from a reflection table list."""
    offsets = calculate_batch_offsets(experiments)
    reflection_tables = assign_batches_to_reflections(reflection_tables, offsets)
    # filter bad refls and negative scales
    batches = flex.int()
    scales = flex.double()
    for r in reflection_tables:
        sel = ~r.get_flags(r.flags.bad_for_scaling, all=False)
        sel &= r["inverse_scale_factor"] > 0
        batches.extend(r["batch"].select(sel))
        scales.extend(r["inverse_scale_factor"].select(sel))
    if not scaled_array:
        scaled_array = scaled_data_as_miller_array(reflection_tables, experiments)
    ms = scaled_array.customized_copy()
    batch_array = miller.array(ms, data=batches)

    batch_ranges = get_batch_ranges(experiments, offsets)
    batch_data = [{"id": i, "range": r} for i, r in enumerate(batch_ranges)]

    properties = batch_dependent_properties(
        batch_array, scaled_array, miller.array(ms, data=scales)
    )

    return properties + (batch_data,)


def rmerge_vs_batch(intensities, batches):
    """Determine batches and Rmerge values per batch."""
    assert intensities.size() == batches.size()
    intensities = intensities.map_to_asu()

    merging = intensities.merge_equivalents()
    merged_intensities = merging.array()

    perm = flex.sort_permutation(batches.data())
    batches = batches.data().select(perm)
    intensities = intensities.select(perm)

    pairs = miller.match_multi_indices(
        merged_intensities.indices(), intensities.indices()
    ).pairs()

    def r_merge_per_batch(pairs):
        """Calculate R_merge for the list of (merged-I, I) pairs."""

        merged_indices, unmerged_indices = zip(*pairs)

        unmerged_Ij = intensities.data().select(flex.size_t(unmerged_indices))
        merged_Ij = merged_intensities.data().select(flex.size_t(merged_indices))

        numerator = flex.sum(flex.abs(unmerged_Ij - merged_Ij))
        denominator = flex.sum(unmerged_Ij)

        if denominator > 0:
            return numerator / denominator
        return 0

    return _batch_bins_and_data(batches, pairs, function_to_apply=r_merge_per_batch)


def i_sig_i_vs_batch(intensities, batches):
    """Determine batches and I/sigma values per batch."""
    assert intensities.size() == batches.size()
    assert intensities.sigmas() is not None
    sel = intensities.sigmas() > 0

    i_sig_i = intensities.data().select(sel) / intensities.sigmas().select(sel)
    batches = batches.select(sel)

    perm = flex.sort_permutation(batches.data())
    batches = batches.data().select(perm)
    i_sig_i = i_sig_i.select(perm)

    return _batch_bins_and_data(batches, i_sig_i, function_to_apply=flex.mean)


def scales_vs_batch(scales, batches):
    """Determine batches and scale values per batch."""
    assert scales.size() == batches.size()

    perm = flex.sort_permutation(batches.data())
    batches = batches.data().select(perm)
    scales = scales.data().select(perm)

    return _batch_bins_and_data(batches, scales, function_to_apply=flex.mean)


def _batch_bins_and_data(batches, values, function_to_apply):
    """Apply function to the data from each batch.

    Return the list of the batch bins and the value for each bin.
    """
    batch_bins = []
    data = []

    i_batch_start = 0
    current_batch = flex.min(batches)
    n_ref = batches.size()
    for i_ref in range(n_ref + 1):
        if i_ref == n_ref or batches[i_ref] != current_batch:
            assert batches[i_batch_start:i_ref].all_eq(current_batch)
            values_for_batch = values[i_batch_start:i_ref]
            data.append(function_to_apply(values_for_batch))
            batch_bins.append(current_batch)
            i_batch_start = i_ref
            if i_ref < n_ref:
                current_batch = batches[i_batch_start]
    return batch_bins, data


formats = {
    "High resolution limit": " %7.2f",
    "Low resolution limit": " %7.2f",
    "Completeness": "%7.1f",
    "Multiplicity": "%7.1f",
    "I/sigma": "%7.1f",
    "Rmerge(I)": "%7.3f",
    "Rmerge(I+/-)": "%7.3f",
    "Rmeas(I)": "%7.3f",
    "Rmeas(I+/-)": "%7.3f",
    "Rpim(I)": "%7.3f",
    "Rpim(I+/-)": "%7.3f",
    "CC half": "%7.3f",
    "Wilson B factor": "%7.3f",
    "Partial bias": "%7.3f",
    "Anomalous completeness": "%7.1f",
    "Anomalous multiplicity": "%7.1f",
    "Anomalous correlation": "%7.3f",
    "Anomalous slope": "%7.3f",
    "dF/F": "%7.3f",
    "dI/s(dI)": "%7.3f",
    "Total observations": "%7d",
    "Total unique": "%7d",
}


def format_statistics(statistics, caption=None):
    """Format for printing statistics from data processing"""

    available = list(statistics.keys())

    columns = len(statistics.get("Completeness", [1, 2, 3]))
    if caption:
        result = caption.ljust(44)
    else:
        result = "".ljust(44)
    if columns == 3:
        result += " Overall    Low     High"
    elif columns == 4:
        result += "Suggested   Low    High  Overall"
    result += "\n"

    for k, format_str in formats.items():
        if k in available:
            try:
                row_data = statistics[k]
                if (
                    columns == 4 and len(row_data) == 1
                ):  # place value in leftmost overall/suggest column
                    row_data = row_data + [None] * (columns - 1)
                row_format = [format_str] + [format_str.strip()] * (len(row_data) - 1)
                formatted = " ".join(
                    (f % k) if k is not None else (" " * len(f % 0))
                    for f, k in zip(row_format, row_data)
                )
            except TypeError:
                formatted = "(error)"
            result += k.ljust(44) + formatted + "\n"

    return result


def table_1_summary(
    merging_statistics,
    anomalous_statistics=None,
    selected_statistics=None,
    selected_anomalous_statistics=None,
    Wilson_B_iso=None,
):
    """Make a summary table of the merging statistics."""
    text = "\n            -------------Summary of merging statistics--------------           \n\n"
    stats = table_1_stats(
        merging_statistics,
        anomalous_statistics,
        selected_statistics,
        selected_anomalous_statistics,
        Wilson_B_iso,
    )
    text += format_statistics(stats)
    return text


def table_1_stats(
    merging_statistics,
    anomalous_statistics=None,
    selected_statistics=None,
    selected_anomalous_statistics=None,
    Wilson_B_iso=None,
):
    """Extract a statistics dict from merging stats objects"""
    key_to_var = {
        "I/sigma": "i_over_sigma_mean",
        "Completeness": "completeness",
        "Low resolution limit": "d_max",
        "Multiplicity": "mean_redundancy",
        "Rmerge(I)": "r_merge",
        "Rmeas(I)": "r_meas",
        "High resolution limit": "d_min",
        "Total observations": "n_obs",
        "Rpim(I)": "r_pim",
        "CC half": "cc_one_half",
        "Total unique": "n_uniq",
    }

    anom_key_to_var = {
        "Rmerge(I+/-)": "r_merge",
        "Rpim(I+/-)": "r_pim",
        "Rmeas(I+/-)": "r_meas",
        "Anomalous completeness": "anom_completeness",
        "Anomalous correlation": "anom_half_corr",
        "Anomalous multiplicity": "mean_redundancy",
    }

    four_column_output = bool(selected_statistics)

    stats = {}

    def generate_stats(d, r, s):
        for key, value in d.items():
            if four_column_output:
                values = (
                    getattr(s.overall, value),
                    getattr(s.bins[0], value),
                    getattr(s.bins[-1], value),
                    getattr(r.overall, value),
                )
            else:
                values = (
                    getattr(r.overall, value),
                    getattr(r.bins[0], value),
                    getattr(r.bins[-1], value),
                )
            if "completeness" in value:
                values = [v_ * 100 for v_ in values]
            if values[0] is not None:
                stats[key] = values

    generate_stats(key_to_var, merging_statistics, selected_statistics)
    if Wilson_B_iso:
        stats["Wilson B factor"] = [Wilson_B_iso]

    if anomalous_statistics:
        anom_probability_plot = (
            anomalous_statistics.overall.anom_probability_plot_expected_delta
        )
        if anom_probability_plot is not None:
            stats["Anomalous slope"] = [anom_probability_plot.slope]
        stats["dF/F"] = [anomalous_statistics.overall.anom_signal]
        stats["dI/s(dI)"] = [
            anomalous_statistics.overall.delta_i_mean_over_sig_delta_i_mean
        ]
        if selected_anomalous_statistics:
            anom_probability_plot = (
                selected_anomalous_statistics.overall.anom_probability_plot_expected_delta
            )
            if anom_probability_plot is not None:
                stats["Anomalous slope"] = [anom_probability_plot.slope]
            stats["dF/F"] = [selected_anomalous_statistics.overall.anom_signal]
            stats["dI/s(dI)"] = [
                selected_anomalous_statistics.overall.delta_i_mean_over_sig_delta_i_mean
            ]
        generate_stats(
            anom_key_to_var, anomalous_statistics, selected_anomalous_statistics
        )

    return stats


def make_merging_statistics_summary(dataset_statistics):
    """Format merging statistics information into an output string."""

    text = "\n            ----------Merging statistics by resolution bin----------           \n\n"
    text += (
        " d_max  d_min   #obs  #uniq   mult.  %comp       <I>  <I/sI>"
        + "    r_mrg   r_meas    r_pim   r_anom   cc1/2   cc_ano\n"
    )
    for bin_stats in dataset_statistics.bins:
        text += bin_stats.format() + "\n"
    text += dataset_statistics.overall.format() + "\n\n"

    return text



"""
Tests for the reindex_to_reference module.
"""

from __future__ import annotations

import pytest

from cctbx import sgtbx

from dials.algorithms.symmetry.cosym._generate_test_data import generate_test_data
from dials.algorithms.symmetry.reindex_to_reference import (
    determine_reindex_operator_against_reference,
)
from dials.util import Sorry


def test_determine_reindex_operator_against_reference():
    """Test that the correct reindex operator is returned by the function."""

    # create a test dataset of random intensities
    data, _ = generate_test_data(
        space_group=sgtbx.space_group_info(symbol="P4").group(), sample_size=1
    )

    # reindexing operator is a,-b,-c
    op = "a,-b,-c"
    # reindex the data into a new array, so that the function should determine
    # that the same change of basis operator should be applied to give consistent
    # indexing back to the original data.
    reindexed_data = data[0].change_basis(op)

    cb_op = determine_reindex_operator_against_reference(data[0], reindexed_data)
    assert cb_op.as_abc() == "a,-b,-c"

    # Repeat but with no reindexing
    cb_op = determine_reindex_operator_against_reference(data[0], data[0])
    assert cb_op.as_abc() == "a,b,c"

    # Test that a Sorry is raised if inconsistent indexing
    data_2, _ = generate_test_data(
        space_group=sgtbx.space_group_info(symbol="P1").group(), sample_size=1
    )
    with pytest.raises(Sorry):
        cb_op = determine_reindex_operator_against_reference(data[0], data_2[0])

    # Test case for a simple space group with no ambiguity
    data, _ = generate_test_data(
        space_group=sgtbx.space_group_info(symbol="P1").group(), sample_size=1
    )
    cb_op = determine_reindex_operator_against_reference(data[0], data[0])
    assert cb_op.as_abc() == "a,b,c"



from __future__ import annotations

import json
import math
import subprocess

import procrunner
import pytest

import scitbx.matrix
from cctbx import sgtbx, uctbx
from dxtbx.model import Crystal, Experiment, ExperimentList, Scan
from dxtbx.serialize import load

from dials.algorithms.symmetry.cosym._generate_test_data import (
    generate_experiments_reflections,
)
from dials.array_family import flex
from dials.command_line import symmetry
from dials.command_line.symmetry import (
    apply_change_of_basis_ops,
    change_of_basis_ops_to_minimum_cell,
    eliminate_sys_absent,
    get_subset_for_symmetry,
    median_unit_cell,
)
from dials.util.exclude_images import exclude_image_ranges_from_scans
from dials.util.multi_dataset_handling import assign_unique_identifiers
from dials.util.phil import parse


def test_symmetry_laue_only(dials_data, tmp_path):
    """Simple test to check that dials.symmetry completes"""

    lcyst_data = dials_data("l_cysteine_dials_output", pathlib=True)
    result = procrunner.run(
        [
            "dials.symmetry",
            lcyst_data / "20_integrated_experiments.json",
            lcyst_data / "20_integrated.pickle",
            lcyst_data / "25_integrated_experiments.json",
            lcyst_data / "25_integrated.pickle",
            "systematic_absences.check=False",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("symmetrized.refl").is_file()
    assert tmp_path.joinpath("symmetrized.expt").is_file()
    exps = load.experiment_list(tmp_path / "symmetrized.expt", check_format=False)
    assert str(exps[0].crystal.get_space_group().info()) == "P 2 2 2"
    joint_reflections = flex.reflection_table.from_file(tmp_path / "symmetrized.refl")
    # check that there are 2 unique id and imageset_ids, and that these
    # correctly correspond to each experiment
    assert len(set(joint_reflections["id"])) == 2
    assert len(set(joint_reflections["imageset_id"])) == 2
    for id_ in range(2):
        sel = joint_reflections["id"] == id_
        assert set(joint_reflections["imageset_id"].select(sel)) == {id_}


def test_symmetry_basis_changes_for_C2(tmp_path):
    """Test the correctness of change of basis operations in dials.symmetry

    Supply the unit cell of beta-lactamase, which triggers a change of
    basis from input to minimum during symmetry analysis."""
    unit_cell = (53.173, 61.245, 69.292, 90.0, 93.04675, 90.0)
    space_group = sgtbx.space_group_info("C 2").group()
    experiments, reflections, _ = generate_experiments_reflections(
        space_group=space_group,
        unit_cell=unit_cell,
        sample_size=1,
        map_to_minimum=False,
    )
    experiments.as_json(tmp_path / "tmp.expt")
    joint_table = flex.reflection_table()
    for r in reflections:
        joint_table.extend(r)
    joint_table.as_file(tmp_path / "tmp.refl")

    result = procrunner.run(
        [
            "dials.symmetry",
            tmp_path / "tmp.expt",
            tmp_path / "tmp.refl",
            "json=symmetry.json",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("symmetrized.refl").is_file()
    symmetrized_expt_file = tmp_path / "symmetrized.expt"
    assert symmetrized_expt_file.is_file()

    expts = load.experiment_list(symmetrized_expt_file, check_format=False)
    for v, expected in zip(expts[0].crystal.get_unit_cell().parameters(), unit_cell):
        assert v == pytest.approx(expected)

    # Using the change of basis ops from the json output we should be able to
    # reindex the input experiments to match the output experiments
    with tmp_path.joinpath("symmetry.json").open() as f:
        d = json.load(f)
    cs = experiments[0].crystal.get_crystal_symmetry()
    cb_op_inp_min = sgtbx.change_of_basis_op(str(d["cb_op_inp_min"][0]))
    cb_op_min_best = sgtbx.change_of_basis_op(str(d["subgroup_scores"][0]["cb_op"]))
    assert cs.change_basis(cb_op_min_best * cb_op_inp_min).is_similar_symmetry(
        expts[0].crystal.get_crystal_symmetry()
    )


@pytest.mark.parametrize("option", ["", "exclude_images=0:1500:1800"])
def test_symmetry_with_absences(dials_data, tmpdir, option):
    """Simple test to check that dials.symmetry, with absences, completes"""

    lcyst_data = dials_data("l_cysteine_dials_output", pathlib=True)
    cmd = [
        "dials.symmetry",
        lcyst_data / "20_integrated_experiments.json",
        lcyst_data / "20_integrated.pickle",
        lcyst_data / "25_integrated_experiments.json",
        lcyst_data / "25_integrated.pickle",
    ]
    if option:
        cmd.append(option)

    result = procrunner.run(cmd, working_directory=tmpdir)
    assert not result.returncode and not result.stderr
    assert tmpdir.join("symmetrized.refl").check()
    assert tmpdir.join("symmetrized.expt").check()
    expts = load.experiment_list(
        tmpdir.join("symmetrized.expt").strpath, check_format=False
    )
    assert str(expts[0].crystal.get_space_group().info()) == "P 21 21 21"


def test_symmetry_with_laue_group_override(dials_data, tmp_path):
    """Simple test to check that dials.symmetry, with overridden laue group, completes"""

    lcyst_data = dials_data("l_cysteine_dials_output", pathlib=True)
    result = procrunner.run(
        [
            "dials.symmetry",
            "laue_group=P121",
            "change_of_basis_op=-b,-a,-c",
            lcyst_data / "20_integrated_experiments.json",
            lcyst_data / "20_integrated.pickle",
            lcyst_data / "25_integrated_experiments.json",
            lcyst_data / "25_integrated.pickle",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("symmetrized.refl").is_file()
    assert tmp_path.joinpath("symmetrized.expt").is_file()
    expts = load.experiment_list(tmp_path / "symmetrized.expt", check_format=False)
    assert str(expts[0].crystal.get_space_group().info()) == "P 1 21 1"
    # Verify that the unit cell has been reindexed correctly
    assert expts[0].crystal.get_unit_cell().parameters() == pytest.approx(
        (8.21578444269, 5.4815363434, 12.1457047712, 90.0, 90.0, 90.0)
    )


@pytest.mark.parametrize("method", ["direct", "fourier"])
def test_symmetry_absences_only(dials_data, tmp_path, method):
    """Test the command line script with real data. Proteinase K in P41"""
    location = dials_data("vmxi_proteinase_k_sweeps", pathlib=True)

    command = [
        "dials.symmetry",
        "laue_group=None",
        location / "experiments_0.json",
        location / "reflections_0.pickle",
        f"method={method}",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("dials.symmetry.html").is_file()
    assert tmp_path.joinpath("symmetrized.expt").is_file()
    exps = load.experiment_list(tmp_path / "symmetrized.expt", check_format=False)
    assert str(exps[0].crystal.get_space_group().info()) == "P 41"

    # Now try with a d_min
    command += ["d_min=4.0"]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    exps = load.experiment_list(tmp_path / "symmetrized.expt", check_format=False)
    assert str(exps[0].crystal.get_space_group().info()) == "P 41"


def test_map_to_minimum_cell():
    # Input and expected output
    input_ucs = [
        (39.7413, 183.767, 140.649, 90, 90, 90),
        (40.16, 142.899, 92.4167, 90, 102.48, 90),
        (180.613, 40.1558, 142.737, 90, 90.0174, 90),
    ]
    input_sgs = ["C 2 2 21", "P 1 2 1", "C 1 2 1"]
    input_hkl = [
        [(1, -75, -71), (1, -73, -70), (1, -71, -69)],
        [(14, -37, -36), (-2, -35, -46), (-3, -34, -47)],
        [(-31, -5, -3), (-25, -3, -3), (-42, -8, -2)],
    ]
    expected_ucs = [
        (39.7413, 94.00755450320204, 140.649, 90.0, 90.0, 77.79717980856927),
        (40.16, 92.46399390642911, 142.899, 90.0, 90.0, 77.3882749092846),
        (
            40.1558,
            92.51154528306184,
            142.73699999999997,
            89.9830147351441,
            90.0,
            77.46527404307477,
        ),
    ]
    expected_output_hkl = [
        [(-1, 37, -71), (-1, 36, -70), (-1, 35, -69)],
        [(-14, 22, 37), (2, 48, 35), (3, 50, 34)],
        [(-5, 13, -3), (-3, 11, -3), (-8, 17, -2)],
    ]

    # Setup the input experiments and reflection tables
    expts = ExperimentList()
    reflections = []
    for uc, sg, hkl in zip(input_ucs, input_sgs, input_hkl):
        uc = uctbx.unit_cell(uc)
        sg = sgtbx.space_group_info(sg).group()
        B = scitbx.matrix.sqr(uc.fractionalization_matrix()).transpose()
        expts.append(Experiment(crystal=Crystal(B, space_group=sg, reciprocal=True)))
        refl = flex.reflection_table()
        refl["miller_index"] = flex.miller_index(hkl)
        reflections.append(refl)

    # Actually run the method we are testing
    cb_ops = change_of_basis_ops_to_minimum_cell(
        expts, max_delta=5, relative_length_tolerance=0.05, absolute_angle_tolerance=2
    )
    cb_ops_as_xyz = [cb_op.as_xyz() for cb_op in cb_ops]
    # Actual cb_ops are machine dependent (sigh)
    assert cb_ops_as_xyz == [
        "-x+y,-2*y,z",
        "-x+z,-z,-y",
        "x+y,-2*x,z",
    ] or cb_ops_as_xyz == ["x-y,2*y,z", "x-z,z,-y", "-x-y,2*x,z"]

    expts_min, reflections = apply_change_of_basis_ops(expts, reflections, cb_ops)
    # Verify that the unit cells have been transformed as expected
    for expt, uc in zip(expts, expected_ucs):
        assert expt.crystal.get_unit_cell().parameters() == pytest.approx(uc, abs=4e-2)

    # Space group should be set to P1
    assert [expt.crystal.get_space_group().type().number() for expt in expts_min] == [
        1,
        1,
        1,
    ]

    # Verify that the reflections have been reindexed as expected
    # Because the exact choice of minimum cell can be platform-dependent,
    # compare the magnitude, but not the sign of the output hkl values
    for refl, expected_hkl in zip(reflections, expected_output_hkl):
        for hkl, e_hkl in zip(refl["miller_index"], expected_hkl):
            assert [abs(h) for h in hkl] == [abs(eh) for eh in e_hkl]


def _make_input_for_exclude_tests(exclude_images=True):
    """Generate input data, that upon exclusion should leave only the first
    reflection."""
    if exclude_images:
        exclude_images = [["0:360:720"], ["1:360:720"]]
    expt1 = Experiment(scan=Scan(image_range=(0, 720), oscillation=(0.0, 1.0)))
    expt2 = Experiment(scan=Scan(image_range=(0, 720), oscillation=(0.0, -1.0)))
    refls1 = flex.reflection_table()
    refls2 = flex.reflection_table()
    refls1["xyzobs.mm.value"] = flex.vec3_double(
        [(0.0, 0.0, 10.0 * math.pi / 180.0), (0.0, 0.0, 370.0 * math.pi / 180.0)]
    )
    refls1["xyzobs.px.value"] = flex.vec3_double([(0.0, 0.0, 10.0), (0.0, 0.0, 370.0)])
    refls1["i"] = flex.int([0, 1])
    refls2["xyzobs.mm.value"] = flex.vec3_double(
        [(0.0, 0.0, -10.0 * math.pi / 180.0), (0.0, 0.0, -370.0 * math.pi / 180.0)]
    )
    refls2["xyzobs.px.value"] = flex.vec3_double([(0.0, 0.0, 10.0), (0.0, 0.0, 370.0)])
    refls2["i"] = flex.int([0, 1])
    expts = ExperimentList([expt1, expt2])
    tables = [refls1, refls2]
    expts, tables = assign_unique_identifiers(expts, tables)
    return exclude_images, expts, tables


def test_get_subset_for_symmetry_image_range():
    """Test that first 360 degrees are selected from each sweep with an exclude
    images command."""
    exclude_images, expts, tables = _make_input_for_exclude_tests(exclude_images=True)
    refls = get_subset_for_symmetry(expts, tables, exclude_images)
    assert refls[0]["i"].all_eq(0)
    assert refls[1]["i"].all_eq(0)


def test_get_subset_for_symmetry_prior_image_range():
    """Test that first 360 degrees are selected from each sweep with an exclude
    images command."""
    exclude_images, expts, tables = _make_input_for_exclude_tests(exclude_images=True)

    # Explicitly exclude a different image range
    expts = exclude_image_ranges_from_scans(tables, expts, [["0:1:360"], ["1:1:360"]])
    refls = get_subset_for_symmetry(expts, tables)
    assert refls[0]["i"].all_eq(1)
    assert refls[1]["i"].all_eq(1)

    # Exclude_images should be cumulative, i.e. the range exclude above should
    # be excluded in addition to the new range provided explicitly to the
    # function
    refls = get_subset_for_symmetry(expts, tables, exclude_images)
    assert len(refls[0]) == 0
    assert len(refls[1]) == 0


def test_get_subset_for_symmetry_first_360():
    """Test that first 360 degrees are selected from each sweep"""
    exclude_images, expts, tables = _make_input_for_exclude_tests(exclude_images=False)
    refls = get_subset_for_symmetry(expts, tables, exclude_images)
    assert refls[0]["i"].all_eq(0)
    assert refls[1]["i"].all_eq(0)


def test_change_of_basis_ops_to_minimum_cell_1037(mocker):
    # See https://github.com/dials/dials/issues/1037

    input_ucs = [
        (
            4.805202948916906,
            12.808064769657364,
            16.544899201125446,
            106.45808502003258,
            90.0065567098825,
            100.77735674275475,
        ),
        (
            4.808011343212577,
            12.821894835790472,
            16.557339561965573,
            106.48431244651402,
            90.0252848479048,
            100.77252933676507,
        ),
        (
            4.8096632137789985,
            12.815648858527567,
            16.55931712239122,
            106.48990701341536,
            90.01703141314147,
            100.80397887485773,
        ),
        (
            4.807294085194974,
            12.822386757910516,
            16.560411742466663,
            106.43185845358086,
            90.02067929544215,
            100.79522302759383,
        ),
    ]

    # Setup the input experiments and reflection tables
    expts = ExperimentList()
    for uc in input_ucs:
        uc = uctbx.unit_cell(uc)
        sg = sgtbx.space_group_info("P1").group()
        B = scitbx.matrix.sqr(uc.fractionalization_matrix()).transpose()
        expts.append(Experiment(crystal=Crystal(B, space_group=sg, reciprocal=True)))

    # We want to spy on the return value of this function
    mocker.spy(symmetry, "unit_cells_are_similar_to")

    # Actually run the method we are testing
    cb_ops = change_of_basis_ops_to_minimum_cell(
        expts, max_delta=5, relative_length_tolerance=0.05, absolute_angle_tolerance=2
    )
    import pytest_mock

    if getattr(pytest_mock, "version", "").startswith("1."):
        assert symmetry.unit_cells_are_similar_to.return_value is True
    else:
        assert symmetry.unit_cells_are_similar_to.spy_return is True
    cb_ops_as_xyz = [cb_op.as_xyz() for cb_op in cb_ops]
    assert len(set(cb_ops_as_xyz)) == 1
    # Actual cb_ops are machine dependent (sigh)
    assert cb_ops_as_xyz[0] in ("x,y,z", "-x,y,-z", "x-y,-y,-z")


def test_change_of_basis_ops_to_minimum_cell_mpro():
    input_ucs = [
        (46.023, 55.001, 64.452, 64.744, 78.659, 89.824),
        (44.747, 53.916, 62.554, 114.985, 99.610, 90.736),
    ]

    # Setup the input experiments and reflection tables
    expts = ExperimentList()
    for uc in input_ucs:
        uc = uctbx.unit_cell(uc)
        sg = sgtbx.space_group_info("P1").group()
        B = scitbx.matrix.sqr(uc.fractionalization_matrix()).transpose()
        expts.append(Experiment(crystal=Crystal(B, space_group=sg, reciprocal=True)))

    # Actually run the method we are testing
    cb_ops = change_of_basis_ops_to_minimum_cell(
        expts, max_delta=5, relative_length_tolerance=0.05, absolute_angle_tolerance=2
    )
    expts.change_basis(cb_ops, in_place=True)
    assert symmetry.unit_cells_are_similar_to(
        expts,
        median_unit_cell(expts),
        relative_length_tolerance=0.05,
        absolute_angle_tolerance=2,
    )


from cctbx import crystal


def test_change_of_basis_ops_to_minimum_cell_with_outlier():
    symmetries = [
        crystal.symmetry(unit_cell=uc, space_group="P1")
        for uc in (
            (52.8868, 52.8868, 333.522, 90, 90, 120),
            (52.6503, 53.0292, 333.783, 89.9872, 89.2247, 60.8078),
            (52.9571, 53.0005, 334.255, 90.0493, 90.0042, 119.893),
            (
                54.4465,
                56.5677,
                355.775,
                93.4376,
                90.0999,
                118.256,
            ),  # This is an outlier
            (52.9235, 52.9235, 335.296, 90, 90, 120),
            (53.4531, 53.4531, 322.909, 90, 90, 120),
        )
    ]

    # Setup the input experiments and reflection tables
    expts = ExperimentList()
    for cs in symmetries:
        B = scitbx.matrix.sqr(cs.unit_cell().fractionalization_matrix()).transpose()
        expts.append(
            Experiment(
                crystal=Crystal(B, space_group=cs.space_group(), reciprocal=True)
            )
        )

    # Actually run the method we are testing
    cb_ops = change_of_basis_ops_to_minimum_cell(
        expts, max_delta=5, relative_length_tolerance=0.05, absolute_angle_tolerance=2
    )
    assert cb_ops.count(None) == 1
    assert cb_ops[3] is None
    expts = ExperimentList([expt for expt, cb_op in zip(expts, cb_ops) if cb_op])
    cb_ops = [cb_op for cb_op in cb_ops if cb_op]

    expts.change_basis(cb_ops, in_place=True)
    assert symmetry.unit_cells_are_similar_to(
        expts,
        median_unit_cell(expts),
        relative_length_tolerance=0.05,
        absolute_angle_tolerance=2,
    )


def test_median_cell():
    unit_cells = [
        uctbx.unit_cell(uc)
        for uc in [
            (10, 11, 11.9, 90, 85, 90),
            (10.1, 11.2, 12, 90, 85.5, 90),
            (10.2, 11.1, 12, 90, 84.7, 90),
        ]
    ]
    expts = ExperimentList()
    for uc in unit_cells:
        sg = sgtbx.space_group_info("P1").group()
        B = scitbx.matrix.sqr(uc.fractionalization_matrix()).transpose()
        expts.append(Experiment(crystal=Crystal(B, space_group=sg, reciprocal=True)))

    median = median_unit_cell(expts)
    assert median.parameters() == pytest.approx((10.1, 11.1, 12, 90, 85, 90))


def test_eliminate_sys_absent():
    refl = flex.reflection_table()
    refl["miller_index"] = flex.miller_index(
        [(-31, -5, -3), (-25, -3, -3), (0, 1, 0), (-42, -8, -2)]
    )
    sgi = sgtbx.space_group_info("C121")
    uc = sgi.any_compatible_unit_cell(volume=1000)
    B = scitbx.matrix.sqr(uc.fractionalization_matrix()).transpose()
    expt = Experiment(crystal=Crystal(B, space_group=sgi.group(), reciprocal=True))
    reflections = eliminate_sys_absent([expt], [refl])
    assert list(reflections[0]["miller_index"]) == [
        (-31, -5, -3),
        (-25, -3, -3),
        (-42, -8, -2),
    ]


def test_few_reflections(dials_data, run_in_tmp_path):
    """
    Test that dials.symmetry does something sensible if given few reflections.

    Use some example integrated data generated from a ten-image 1° sweep.  These
    contain a few dozen integrated reflections.

    Args:
        dials_data: DIALS custom Pytest fixture for access to test data.
        run_in_tmpdir: DIALS custom Pytest fixture to run this test in a temporary
                       directory.
    """
    # Get and use the default parameters for dials.symmetry.
    params = symmetry.phil_scope.fetch(source=parse("")).extract()

    # Use the integrated data from the first ten images of the first sweep.
    data_dir = dials_data("l_cysteine_dials_output", pathlib=True)
    experiments = ExperimentList.from_file(data_dir / "11_integrated.expt")
    reflections = [flex.reflection_table.from_file(data_dir / "11_integrated.refl")]

    # Run dials.symmetry on the above data files.
    symmetry.symmetry(experiments, reflections, params)


def test_x4wide(dials_data, tmp_path):
    """Verify reflections are reindexed before systematic absence analysis.

    Exercise a bug fix in https://github.com/dials/dials/pull/2183

    Expected space group is P 43 21 2 (or its enantiomorphic equivalent, P 41 21 2)
    See also https://www.rcsb.org/structure/3QF8
    """
    x4wide = dials_data("x4wide_processed", pathlib=True)
    result = subprocess.run(
        [
            "dials.symmetry",
            x4wide / "AUTOMATIC_DEFAULT_scaled.expt",
            x4wide / "AUTOMATIC_DEFAULT_scaled.refl",
        ],
        cwd=tmp_path,
    )
    assert not result.returncode and not result.stderr
    assert tmp_path.joinpath("symmetrized.refl").is_file()
    assert tmp_path.joinpath("symmetrized.expt").is_file()
    exps = load.experiment_list(tmp_path / "symmetrized.expt", check_format=False)
    assert str(exps[0].crystal.get_space_group().info()) == "P 41 21 2"



from __future__ import annotations

import logging
import sys

from libtbx.phil import parse

from dials.algorithms.statistics.cc_half_algorithm import CCHalfFromDials, CCHalfFromMTZ
from dials.util import log, show_mail_handle_errors
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files

logger = logging.getLogger("dials.command_line.compute_delta_cchalf")

help_message = """

This program computes the delta cchalf excluding images
"""

# Set the phil scope
phil_scope = parse(
    """

  input {

    mtzfile = None
      .type = str
      .help = "We can also import an MTZ file"

  }

  mode = *dataset image_group
    .type = choice
    .help = "Perform analysis on whole datasets or batch groups"

  group_size = 10
    .type = int(value_min=1)
    .help = "The number of images to group together when calculating delta"
            "cchalf in image_group mode"

  mtz {
    batch_offset = None
      .type = int
      .help = "The batch offset between consecutive datasets in the mtz file."
  }

  output {

    experiments = "filtered.expt"
      .type = str
      .help = "The filtered experiments file"

    reflections = "filtered.refl"
      .type = str
      .help = "The filtered reflections file"

    table = "delta_cchalf.dat"
      .type = str
      .help = "A file with delta cchalf values"
    html = "compute_delta_cchalf.html"
      .type = str
      .help = "HTML filename for report of results."
  }

  nbins = 10
    .type = int(value_min=1)
    .help = "The number of resolution bins to use"

  dmin = None
    .type = float
    .help = "The maximum resolution"

  dmax = None
    .type = float
    .help = "The minimum resolution"

  stdcutoff = 4.0
    .type = float
    .help = "Datasets with a ΔCC½ below (mean - stdcutoff*std) are removed"

  output {
    log = 'dials.compute_delta_cchalf.log'
      .type = str
      .help = "The log filename"
  }
"""
)


@show_mail_handle_errors()
def run(args=None, phil=phil_scope):
    """Run the command-line script."""

    usage = "dials.compute_delta_cchalf [options] scaled.expt scaled.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        epilog=help_message,
        read_experiments=True,
        read_reflections=True,
        check_format=False,
    )

    params, _ = parser.parse_args(args=args, show_diff_phil=False)

    log.config(logfile=params.output.log)

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if not experiments and not reflections:
        if not params.input.mtzfile:
            parser.print_help()
            return
        else:
            try:
                script = CCHalfFromMTZ(params, params.input.mtzfile)
            except ValueError as e:
                sys.exit(f"Error: {e}")
    else:
        if not experiments or not reflections:
            parser.print_help()
            return
        else:
            if not len(reflections) == 1:
                exit("Only one reflection table can be provided")
            n_datasets = len(set(reflections[0]["id"]).difference({-1}))
            if n_datasets != len(experiments):
                exit(
                    """
The number of experiments (%s) does not match the number
of datasets in the reflection table (%s)
""",
                    len(experiments),
                    n_datasets,
                )
            try:
                script = CCHalfFromDials(params, experiments, reflections[0])
            except ValueError as e:
                sys.exit(f"Error: {e}")

    script.run()
    script.output()


if __name__ == "__main__":
    run()



from __future__ import annotations

import logging

from dials_algorithms_background_modeller_ext import (
    BackgroundStatistics,
    MultiPanelBackgroundStatistics,
)

__all__ = [
    "BackgroundModeller",
    "BackgroundModellerExecutor",
    "BackgroundModellerResult",
    "BackgroundStatistics",
    "FinalizeModel",
    "MultiPanelBackgroundStatistics",
]

logger = logging.getLogger(__name__)


class FinalizeModel:
    """
    A class to finalize the background model
    """

    def __init__(self, experiments, filter_type="median", kernel_size=10, niter=100):
        """
        Initialize the finalizer

        :param experiments: The experiment list
        :param kernel_size: The median filter kernel size
        :param niter: The number of iterations for filling holes
        """
        from dials.algorithms.background.gmodel import PolarTransform

        # Set some parameters
        self.filter_type = filter_type
        self.kernel_size = kernel_size
        self.niter = niter

        # Check the input
        assert len(experiments) == 1
        experiment = experiments[0]
        assert len(experiment.detector) == 1

        # Save the experiment
        self.experiment = experiment

        # Create the transform object
        self.transform = PolarTransform(
            experiment.beam, experiment.detector[0], experiment.goniometer
        )

    def finalize(self, data, mask):
        """
        Finalize the model

        :param data: The data array
        :param mask: The mask array
        """
        from dials.algorithms.image.fill_holes import diffusion_fill, simple_fill
        from dials.algorithms.image.filter import mean_filter, median_filter
        from dials.array_family import flex

        # Print some image properties
        sub_data = data.as_1d().select(mask.as_1d())
        logger.info("Raw image statistics:")
        logger.info("  min:  %d", int(flex.min(sub_data)))
        logger.info("  max:  %d", int(flex.max(sub_data)))
        logger.info("  mean: %d", int(flex.mean(sub_data)))
        logger.info("")

        # Transform to polar
        logger.info("Transforming image data to polar grid")
        result = self.transform.to_polar(data, mask)
        data = result.data()
        mask = result.mask()
        sub_data = data.as_1d().select(mask.as_1d())
        logger.info("Polar image statistics:")
        logger.info("  min:  %d", int(flex.min(sub_data)))
        logger.info("  max:  %d", int(flex.max(sub_data)))
        logger.info("  mean: %d", int(flex.mean(sub_data)))
        logger.info("")

        # Filter the image to remove noise
        if self.kernel_size > 0:
            if self.filter_type == "median":
                logger.info("Applying median filter")
                data = median_filter(data, mask, (self.kernel_size, 0), periodic=True)
                sub_data = data.as_1d().select(mask.as_1d())
                logger.info("Median polar image statistics:")
                logger.info("  min:  %d", int(flex.min(sub_data)))
                logger.info("  max:  %d", int(flex.max(sub_data)))
                logger.info("  mean: %d", int(flex.mean(sub_data)))
                logger.info("")
            elif self.filter_type == "mean":
                logger.info("Applying mean filter")
                mask_as_int = mask.as_1d().as_int()
                mask_as_int.reshape(mask.accessor())
                data = mean_filter(data, mask_as_int, (self.kernel_size, 0), 1)
                sub_data = data.as_1d().select(mask.as_1d())
                logger.info("Mean polar image statistics:")
                logger.info("  min:  %d", int(flex.min(sub_data)))
                logger.info("  max:  %d", int(flex.max(sub_data)))
                logger.info("  mean: %d", int(flex.mean(sub_data)))
                logger.info("")
            else:
                raise RuntimeError(f"Unknown filter_type: {self.filter_type}")

        # Fill any remaining holes
        logger.info("Filling holes")
        data = simple_fill(data, mask)
        data = diffusion_fill(data, mask, self.niter)
        mask = flex.bool(data.accessor(), True)
        sub_data = data.as_1d().select(mask.as_1d())
        logger.info("Filled polar image statistics:")
        logger.info("  min:  %d", int(flex.min(sub_data)))
        logger.info("  max:  %d", int(flex.max(sub_data)))
        logger.info("  mean: %d", int(flex.mean(sub_data)))
        logger.info("")

        # Transform back
        logger.info("Transforming image data from polar grid")
        result = self.transform.from_polar(data, mask)
        data = result.data()
        mask = result.mask()
        sub_data = data.as_1d().select(mask.as_1d())
        logger.info("Final image statistics:")
        logger.info("  min:  %d", int(flex.min(sub_data)))
        logger.info("  max:  %d", int(flex.max(sub_data)))
        logger.info("  mean: %d", int(flex.mean(sub_data)))
        logger.info("")

        # Fill in any discontinuities
        mask = ~self.transform.discontinuity()[:-1, :-1]
        data = diffusion_fill(data, mask, self.niter)

        # Get and apply the mask
        mask = self.experiment.imageset.get_mask(0)[0]
        mask = mask.as_1d().as_int().as_double()
        mask.reshape(data.accessor())
        data *= mask

        # Return the result
        return data


class BackgroundModellerResult:
    """
    A class to contain the modelling result
    """

    def __init__(
        self,
        mean=None,
        variance=None,
        dispersion=None,
        mask=None,
        min_image=None,
        max_image=None,
        model=None,
        polar_model=None,
    ):
        """
        Init the result
        """
        self.mean = mean
        self.variance = variance
        self.dispersion = dispersion
        self.mask = mask
        self.min_image = min_image
        self.max_image = max_image
        self.model = model
        self.polar_model = polar_model


class BackgroundModellerExecutor:
    def __init__(self, experiments, params):
        assert len(experiments) == 1
        self.min_images = params.modeller.min_images
        if self.min_images > len(experiments[0].imageset):
            self.min_images = len(experiments[0].imageset)
        self.image_type = params.modeller.image_type

        self.finalizer = FinalizeModel(
            experiments=experiments,
            filter_type=params.modeller.filter_type,
            kernel_size=params.modeller.kernel_size,
            niter=params.modeller.niter,
        )
        self.result = None

    def process(self, image_volume, experiments, reflections):
        from dials.algorithms.integration.processor import job

        # Write some output
        logger.info(
            " Background modelling; job: %d; frames: %d -> %d; # Reflections: %d",
            job.index,
            image_volume.frame0(),
            image_volume.frame1(),
            len(reflections),
        )

        # Compute the shoebox mask
        reflections.compute_mask(experiments=experiments, image_volume=image_volume)

        # Compute the sum, sum^2 and the number of contributing pixels
        return MultiPanelBackgroundStatistics(image_volume)

    def accumulate(self, index, data):
        if self.result is None:
            self.result = data
        else:
            self.result += data

    def finalize_model(self):
        logger.info("")
        logger.info("=" * 80)
        logger.info("Finalizing model")
        logger.info("")

        result = []
        for i in range(len(self.result)):

            # Get the statistics
            stats = self.result.get(i)
            mean = stats.mean(self.min_images)
            variance = stats.variance(self.min_images)
            dispersion = stats.dispersion(self.min_images)
            mask = stats.mask(self.min_images)
            min_image = stats.min()
            max_image = stats.max()

            # Create the model
            if self.image_type == "min":
                model = self.finalizer.finalize(min_image, mask)
            elif self.image_type == "mean":
                model = self.finalizer.finalize(mean, mask)
            else:
                raise RuntimeError(f"Unknown image_type: {self.image_type}")

            # Add to the list
            result.append(
                BackgroundModellerResult(
                    mean=mean,
                    variance=variance,
                    dispersion=dispersion,
                    mask=mask,
                    min_image=min_image,
                    max_image=max_image,
                    model=model,
                )
            )

        return result


class BackgroundModeller:
    """
    A class to help with background modelling
    """

    def __init__(self, experiments, reflections, params):
        """
        Initialize the modeller

        :param experiments: The experiment list
        :param reflections: The reflections to process
        :param params: The parameters to use
        """
        # Check all reflections have same imageset and get it
        imageset = experiments[0].imageset
        for expr in experiments:
            assert expr.imageset == imageset, "All experiments must share and imageset"

        # Save some stuff
        self.experiments = experiments
        self.reflections = reflections
        self.params = params
        self.model = None

    def compute(self):
        """
        Integrate the data
        """
        from dials.algorithms.integration.image_integrator import ProcessorImage
        from dials.util.command_line import heading

        # Init the report
        self.profile_model_report = None
        self.integration_report = None

        # Create summary format
        fmt = (
            " Processing the following experiments:\n"
            "\n"
            " Experiments: %d\n"
            " Beams:       %d\n"
            " Detectors:   %d\n"
            " Goniometers: %d\n"
            " Scans:       %d\n"
            " Crystals:    %d\n"
            " Imagesets:   %d\n"
        )

        # Print the summary
        logger.info(
            fmt,
            len(self.experiments),
            len(self.experiments.beams()),
            len(self.experiments.detectors()),
            len(self.experiments.goniometers()),
            len(self.experiments.scans()),
            len(self.experiments.crystals()),
            len(self.experiments.imagesets()),
        )

        # Print a heading
        logger.info("=" * 80)
        logger.info("")
        logger.info(heading("Modelling background"))
        logger.info("")

        # Expand n_sigma
        for expt in self.experiments:
            expt.profile._n_sigma += 2

        # Compute some reflection properties
        self.reflections.compute_zeta_multi(self.experiments)
        self.reflections.compute_d(self.experiments)
        self.reflections.compute_bbox(self.experiments)

        # Construvt the image integrator processor
        processor = ProcessorImage(self.experiments, self.reflections, self.params)
        processor.executor = BackgroundModellerExecutor(self.experiments, self.params)

        # Do the processing
        _, time_info = processor.process()

        # Compute the model
        self.model = processor.executor.finalize_model()

        # Print the time info
        logger.info(str(time_info))
        logger.info("")

        # Return the reflections
        return self.model



from __future__ import annotations

import math

from dials.algorithms.profile_model.modeller import EmpiricalProfileModeller


def evaluate_gaussian(x, a, x0, sx):
    assert len(x) == len(x0)
    assert len(x) == len(sx)

    g = 0.0
    for xi, x0i, sxi in zip(x, x0, sx):
        g += (xi - x0i) ** 2 / (2.0 * sxi**2)

    return a * math.exp(-g)


def gaussian(size, a, x0, sx):
    from scitbx.array_family import flex

    result = flex.double(flex.grid(size))

    index = [0] * len(size)
    while True:
        result[index] = evaluate_gaussian(index, a, x0, sx)
        for j in range(len(size)):
            index[j] += 1
            if index[j] < size[j]:
                break
            index[j] = 0
            if j == len(size) - 1:
                return result


class Modeller(EmpiricalProfileModeller):
    def model(self, reflections, profiles):
        from dials.array_family import flex

        indices = flex.size_t(range(len(self)))
        weights = flex.double([1.0] * len(self))
        for profile in profiles:
            self.add(indices, weights, profile)


class Test:
    def setup_class(self):
        self.n = 9
        self.grid_size = (9, 9, 9)
        self.threshold = 0.0

    def test_with_identical_non_negative_profiles(self):
        from scitbx.array_family import flex

        # Generate identical non-negative profiles
        reflections, profiles, profile = self.generate_identical_non_negative_profiles()

        # Create the reference learner
        modeller = Modeller(self.n, self.grid_size, self.threshold)

        # Do the modelling
        modeller.model(reflections, profiles)
        modeller.finalize()

        # Normalize the profile
        profile = self.normalize_profile(profile)

        # Check that all the reference profiles are the same
        eps = 1e-10
        for index in range(len(modeller)):
            reference = modeller.data(index)
            for k in range(self.grid_size[2]):
                for j in range(self.grid_size[1]):
                    for i in range(self.grid_size[0]):
                        assert abs(reference[k, j, i] - profile[k, j, i]) <= eps
            assert abs(flex.sum(reference) - 1.0) <= eps

    def test_with_systematically_offset_profiles(self):
        from scitbx.array_family import flex

        # Generate identical non-negative profiles
        reflections, profiles = self.generate_systematically_offset_profiles()

        # Create the reference learner
        modeller = Modeller(self.n, self.grid_size, self.threshold)

        # Do the modelling
        modeller.model(reflections, profiles)
        modeller.finalize()

        # Check that all the reference profiles are the same
        eps = 1e-10
        profile = None
        for index in range(len(modeller)):
            reference = modeller.data(index)
            if profile is not None:
                for k in range(self.grid_size[2]):
                    for j in range(self.grid_size[1]):
                        for i in range(self.grid_size[0]):
                            assert abs(reference[k, j, i] - profile[k, j, i]) <= eps
            else:
                profile = reference
            assert abs(flex.sum(reference) - 1.0) <= eps

    def normalize_profile(self, profile):
        from scitbx.array_family import flex

        max_profile = flex.max(profile)
        threshold = self.threshold * max_profile
        sum_profile = 0.0
        for i in range(len(profile)):
            if profile[i] > threshold:
                sum_profile += profile[i]
            else:
                profile[i] = 0.0

        result = flex.double(flex.grid(profile.all()))
        for i in range(len(profile)):
            result[i] = profile[i] / sum_profile

        return result

    def generate_identical_non_negative_profiles(self):
        from random import uniform

        from dials.array_family import flex

        rlist = flex.reflection_table(1000)

        profile = gaussian(self.grid_size, 1000, (4, 4, 4), (1.5, 1.5, 1.5))

        xyz = flex.vec3_double(1000)
        profiles = []
        for i in range(1000):
            x = uniform(0, 1000)
            y = uniform(0, 1000)
            z = uniform(0, 10)
            xyz[i] = (x, y, z)
            profiles.append(profile.deep_copy())
        rlist["xyzcal.px"] = xyz

        return rlist, profiles, profile

    def generate_systematically_offset_profiles(self):
        from random import uniform

        from dials.array_family import flex

        rlist = flex.reflection_table(1000)

        xyz = flex.vec3_double(1000)
        profiles = []
        for i in range(1000):
            x = uniform(0, 1000)
            y = uniform(0, 1000)
            z = uniform(0, 10)

            offset = -4.5 + 9 * x / 1000.0

            profile = gaussian(
                self.grid_size, 1000, (4 + offset, 4, 4), (1.5, 1.5, 1.5)
            )
            xyz[i] = (x, y, z)
            profiles.append(profile)

        rlist["xyzcal.px"] = xyz
        return rlist, profiles



# DIALS_ENABLE_COMMAND_LINE_COMPLETION
"""
This program is used to integrate the reflections on the diffraction images. It
is called with an experiment list outputted from dials.index or dials.refine and
a corresponding set of strong spots from which a profile model is calculated.
The program will output a set of integrated reflections and an experiment list
with additional profile model data. The data can be reintegrated using the same
profile model by inputting this integrated.expt file back into
dials.integate.

Examples::

  dials.integrate models.expt refined.refl

  dials.integrate models.expt refined.refl output.reflections=integrated.refl

  dials.integrate models.expt refined.refl profile.fitting=False

  dials.integrate models.expt refined.refl background.algorithm=glm
"""


from __future__ import annotations

import logging
import math
import sys

from orderedset import OrderedSet

from dxtbx.model.experiment_list import Experiment, ExperimentList
from libtbx.phil import parse

import dials.util.log
from dials.algorithms.integration.integrator import create_integrator
from dials.algorithms.profile_model.factory import ProfileModelFactory
from dials.array_family import flex
from dials.util import show_mail_handle_errors
from dials.util.command_line import heading
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.slice import slice_crystal
from dials.util.version import dials_version

logger = logging.getLogger("dials.command_line.integrate")

# Create the phil scope

phil_scope = parse(
    """

  output {
    experiments = 'integrated.expt'
      .type = str
      .help = "The experiments output filename"

    output_unintegrated_reflections = False
      .type = bool
      .expert_level = 2
      .help = "Include unintegrated reflections in output file"

    reflections = 'integrated.refl'
      .type = str
      .help = "The integrated output filename"

    phil = 'dials.integrate.phil'
      .type = str
      .help = "The output phil file"

    log = 'dials.integrate.log'
      .type = str
      .help = "The log filename"

    report = None
      .type = str
      .help = "The integration report filename (*.xml or *.json)"

    include_bad_reference = False
      .type = bool
      .help = "Include bad reference data including unindexed spots,"
              "and reflections whose predictions are messed up in the"
              "reflection table output. Reflections will have the"
              "'bad_reference' flag set."
  }

  scan_range = None
    .type = ints(size=2)
    .help = "Explicitly specify the images to be processed. Only applicable"
            "when experiment list contains a single imageset."
    .multiple = True

  create_profile_model = True
    .type = bool
    .help = "Create the profile model"

  sampling
    .expert_level = 1
  {

    reflections_per_degree = 50
      .help = "The number of predicted reflections per degree of the sequence "
              "to integrate."
      .type = float(value_min=0.)

    minimum_sample_size = 1000
      .help = "cutoff that determines whether subsetting of the input "
              "prediction list is done"
      .type = int

    maximum_sample_size = None
      .help = "The maximum number of predictions to integrate."
              "Overrides reflections_per_degree if that produces a"
              "larger sample size."
      .type = int(value_min=1)

    integrate_all_reflections = True
      .help = "Override reflections_per_degree and integrate all predicted"
              "reflections."
      .type = bool

    random_seed = 0
      .help = "Random seed for sampling"
      .type = int

  }

  exclude_images = None
    .type = ints
    .help = "Exclude images from integration (e.g. 1,2,3,4,5 etc)"

  include scope dials.algorithms.integration.integrator.phil_scope
  include scope dials.algorithms.profile_model.factory.phil_scope
  include scope dials.algorithms.spot_prediction.reflection_predictor.phil_scope
  include scope dials.algorithms.integration.stills_significance_filter.phil_scope
  include scope dials.algorithms.integration.kapton_correction.absorption_phil_scope
""",
    process_includes=True,
)

# Local overrides for dials.integrate
phil_overrides = parse(
    """
integration {
  mp {
    nproc = Auto
  }
}
"""
)
working_phil = phil_scope.fetch(sources=[phil_overrides])


def process_reference(reference):
    """
    Remove bad reflections from the reference.

    Remove unindexed, bad_for_refinement, bad miller index.

    Args:
        reference: A reflection table.

    Returns:
        (tuple): tuple containing:

            reference: A reduction of the input reference reflection table.
            rubbish: A reflection table containing the reflections filtered out of
                the input table.

    Raises:
        ValueError: If no indexed spots, bad id, unmatched panel.
    """

    if reference is None:
        return None, None
    assert "miller_index" in reference
    assert "id" in reference
    logger.info(
        "Processing reference reflections\n read %d strong spots", reference.size()
    )
    mask = reference.get_flags(reference.flags.indexed)
    rubbish = reference.select(~mask)
    n_unindexed = mask.count(False)
    if n_unindexed > 0:
        reference.del_selected(~mask)
        logger.info(" removing %d unindexed reflections", n_unindexed)
    if reference.size() == 0:
        raise ValueError(
            "Invalid input for reference reflections. No indexed spots found."
        )
    mask = reference.get_flags(reference.flags.bad_for_refinement, all=False)
    n_masked = mask.count(True)
    if n_masked:
        rubbish.extend(reference.select(mask))
        reference.del_selected(mask)
        logger.info(" removing %d reflections marked as bad for refinement", n_masked)
    mask = reference["miller_index"] == (0, 0, 0)
    n_masked = mask.count(True)
    if n_masked > 0:
        rubbish.extend(reference.select(mask))
        reference.del_selected(mask)
        logger.info(" removing %d reflections with hkl (0,0,0)", n_masked)
    mask = reference["id"] < 0
    n_masked = mask.count(True)
    if n_masked > 0:
        raise ValueError(
            """
    Invalid input for reference reflections.
    %d reference spots have an invalid experiment id
    """
            % n_masked
        )
    if (reference["panel"] == reference["shoebox"].panels()).count(False) > 0:
        raise ValueError(
            'reflection table "panel" column does not match "shoebox" panel'
        )
    logger.info(" using %d indexed reflections", reference.size())
    logger.info(" found %d junk reflections", rubbish.size())
    return reference, rubbish


def filter_reference_pixels(reference, experiments):
    """
    Set any pixel closer to other reflections to background.

    Args:
        reference: A reflection table
        experiments: The experiment list

    Returns:
        The input reflection table with modified shoeboxes.
    """
    modified_count = 0
    for experiment, indices in reference.iterate_experiments_and_indices(experiments):
        subset = reference.select(indices)
        modified = subset["shoebox"].mask_neighbouring(
            subset["miller_index"],
            experiment.beam,
            experiment.detector,
            experiment.goniometer,
            experiment.scan,
            experiment.crystal,
        )
        modified_count += modified.count(True)
        reference.set_selected(indices, subset)
    logger.info(" masked neighbouring pixels in %d shoeboxes", modified_count)
    return reference


def sample_predictions(experiments, predicted, params):
    """
    Select a random sample of the predicted reflections to integrate.

    Args:
        experiments: The experiment list
        predicted: A reflection table of predicted reflections
        params: The integration phil parameters

    Returns:
        A subset of the original predicted table.
    """

    if params.sampling.random_seed:
        flex.set_random_seed(params.sampling.random_seed)

    nref_per_degree = params.sampling.reflections_per_degree
    min_sample_size = params.sampling.minimum_sample_size
    max_sample_size = params.sampling.maximum_sample_size

    # this code is very similar to David's code in algorithms/refinement/reflection_manager.py!

    working_isel = flex.size_t()
    for iexp, exp in enumerate(experiments):

        sel = predicted["id"] == iexp
        isel = sel.iselection()
        nrefs = sample_size = len(isel)

        # set sample size according to nref_per_degree (per experiment)
        if exp.scan and nref_per_degree:
            sequence_range_rad = exp.scan.get_oscillation_range(deg=False)
            width = math.degrees(abs(sequence_range_rad[1] - sequence_range_rad[0]))
            sample_size = int(nref_per_degree * width)
        else:
            sequence_range_rad = None

        # adjust sample size if below the chosen limit
        sample_size = max(sample_size, min_sample_size)

        # set maximum sample size if requested
        if max_sample_size:
            sample_size = min(sample_size, max_sample_size)

        # determine subset and collect indices
        if sample_size < nrefs:
            isel = isel.select(flex.random_selection(nrefs, sample_size))
        working_isel.extend(isel)

    # create subset
    return predicted.select(working_isel)


def split_for_scan_range(experiments, reference, scan_range):
    """Update experiments when scan range is set.

    Args:
        experiments: An experiment list
        reference: A reflection table of reference reflections
        scan_range (tuple): Range of scan images to be processed

    Returns:
        experiments: A new experiment list with the requested scan ranges
        reference: A reflection table with data from the scan ranges

    Raises:
        ValueError: If bad input for scan range.
    """

    # Only do anything is the scan range is set
    if scan_range is not None and len(scan_range) > 0:

        # Ensure that all experiments have the same imageset and scan
        iset = [e.imageset for e in experiments]
        scan = [e.scan for e in experiments]
        assert all(x == iset[0] for x in iset)
        assert all(x == scan[0] for x in scan)

        # Get the imageset and scan
        iset = experiments[0].imageset
        scan = experiments[0].scan

        # Get the array range
        if scan is not None:
            frames_start, frames_end = scan.get_array_range()
            assert scan.get_num_images() == len(iset)
        else:
            frames_start, frames_end = (0, len(iset))

        # Create the new lists
        new_experiments = ExperimentList()
        new_reference_all = reference.split_by_experiment_id()
        new_reference = flex.reflection_table()
        for i in range(len(new_reference_all) - len(experiments)):
            new_reference_all.append(flex.reflection_table())
        assert len(new_reference_all) == len(experiments)

        # Loop through all the scan ranges and create a new experiment list with
        # the requested scan ranges.
        for scan_start, scan_end in scan_range:
            # Validate the requested scan range
            if scan_end == scan_start:
                raise ValueError(
                    "Scan range end must be higher than start; pass {},{} for single image".format(
                        scan_start, scan_start + 1
                    )
                )
            if scan_end < scan_start:
                raise ValueError("Scan range must be in ascending order")
            elif scan_start < frames_start or scan_end > frames_end:
                raise ValueError(
                    "Scan range must be within image range {}..{}".format(
                        frames_start, frames_end
                    )
                )

            assert scan_end > scan_start
            assert scan_start >= frames_start
            assert scan_end <= frames_end

            index_start = scan_start - frames_start
            index_end = index_start + (scan_end - scan_start)
            assert index_start < index_end
            assert index_start >= 0
            assert index_end <= len(iset)
            new_iset = iset[index_start:index_end]
            if scan is None:
                new_scan = None
            else:
                new_scan = scan[index_start:index_end]

            for i, e1 in enumerate(experiments):
                e2 = Experiment()
                e2.beam = e1.beam
                e2.detector = e1.detector
                e2.goniometer = e1.goniometer
                e2.crystal = slice_crystal(e1.crystal, (index_start, index_end))
                e2.profile = e1.profile
                e2.imageset = new_iset
                e2.scan = new_scan
                new_reference_all[i]["id"] = flex.int(
                    len(new_reference_all[i]), len(new_experiments)
                )
                new_reference.extend(new_reference_all[i])
                new_experiments.append(e2)
        experiments = new_experiments
        reference = new_reference

        # Print some information
        logger.info("Modified experiment list to integrate over requested scan range")
        for scan_start, scan_end in scan_range:
            logger.info(" scan_range = %d -> %d", scan_start, scan_end)

    # Return the experiments
    return experiments, reference


def run_integration(params, experiments, reference=None):
    """Perform the integration.

    Returns:
        experiments: The integrated experiments
        reflections: The integrated reflections
        report(optional): An integration report.

    Raises:
        ValueError: For a number of bad inputs
        RuntimeError: If the profile model creation fails
    """
    predicted = None
    rubbish = None

    for abs_params in params.absorption_correction:
        if abs_params.apply:
            if not (
                params.integration.debug.output
                and not params.integration.debug.separate_files
            ):
                raise ValueError(
                    "Shoeboxes must be saved to integration intermediates to apply an absorption correction. "
                    + "Set integration.debug.output=True, integration.debug.separate_files=False and "
                    + "integration.debug.delete_shoeboxes=True to temporarily store shoeboxes."
                )

    # Print if we're using a mask
    for i, exp in enumerate(experiments):
        mask = exp.imageset.external_lookup.mask
        if mask.filename is not None:
            if mask.data:
                logger.info("Using external mask: %s", mask.filename)
                for tile in mask.data:
                    logger.info(" Mask has %d pixels masked", tile.data().count(False))

    # Print the experimental models
    for i, exp in enumerate(experiments):
        summary = "\n".join(
            (
                "",
                "=" * 80,
                "",
                "Experiments",
                "",
                "Models for experiment %d" % i,
                "",
                str(exp.beam),
                str(exp.detector),
            )
        )
        if exp.goniometer:
            summary += str(exp.goniometer) + "\n"
        if exp.scan:
            summary += str(exp.scan) + "\n"
        summary += str(exp.crystal)
        logger.info(summary)

    logger.info("\n".join(("", "=" * 80, "")))
    logger.info(heading("Initialising"))

    # Load the data
    if reference:
        reference, rubbish = process_reference(reference)

        # Check pixels don't belong to neighbours
        if exp.goniometer is not None and exp.scan is not None:
            reference = filter_reference_pixels(reference, experiments)

        # Modify experiment list if scan range is set.
        experiments, reference = split_for_scan_range(
            experiments, reference, params.scan_range
        )

    # Modify experiment list if exclude images is set
    if params.exclude_images:
        for experiment in experiments:
            for index in params.exclude_images:
                experiment.imageset.mark_for_rejection(index, True)

    # Predict the reflections
    logger.info("\n".join(("", "=" * 80, "")))
    logger.info(heading("Predicting reflections"))
    predicted = flex.reflection_table.from_predictions_multi(
        experiments,
        dmin=params.prediction.d_min,
        dmax=params.prediction.d_max,
        margin=params.prediction.margin,
        force_static=params.prediction.force_static,
        padding=params.prediction.padding,
    )
    isets = OrderedSet(e.imageset for e in experiments)
    predicted["imageset_id"] = flex.int(predicted.size(), 0)
    if len(isets) > 1:
        for e in experiments:
            iset_id = isets.index(e.imageset)
            for id_ in predicted.experiment_identifiers().keys():
                identifier = predicted.experiment_identifiers()[id_]
                if identifier == e.identifier:
                    sel = predicted["id"] == id_
                    predicted["imageset_id"].set_selected(sel, iset_id)
                    break

    # Match reference with predicted
    if reference:
        matched, reference, unmatched = predicted.match_with_reference(reference)
        assert len(matched) == len(predicted)
        assert matched.count(True) <= len(reference)
        if matched.count(True) == 0:
            raise ValueError(
                """
        Invalid input for reference reflections.
        Zero reference spots were matched to predictions
    """
            )
        elif unmatched:
            msg = (
                "Warning: %d reference spots were not matched to predictions"
                % unmatched.size()
            )
            border = "\n".join(("", "*" * 80, ""))
            logger.info("".join((border, msg, border)))
            rubbish.extend(unmatched)

        if len(experiments) > 1:
            # filter out any experiments without matched reference reflections
            # f_: filtered

            f_reference = flex.reflection_table()
            f_predicted = flex.reflection_table()
            f_rubbish = flex.reflection_table()
            f_experiments = ExperimentList()
            good_expt_count = 0

            def refl_extend(src, dest, eid):
                old_id = eid
                new_id = good_expt_count
                tmp = src.select(src["id"] == old_id)
                tmp["id"] = flex.int(len(tmp), good_expt_count)
                if old_id in tmp.experiment_identifiers():
                    identifier = tmp.experiment_identifiers()[old_id]
                    del tmp.experiment_identifiers()[old_id]
                    tmp.experiment_identifiers()[new_id] = identifier
                dest.extend(tmp)

            for expt_id, experiment in enumerate(experiments):
                if len(reference.select(reference["id"] == expt_id)) != 0:
                    refl_extend(reference, f_reference, expt_id)
                    refl_extend(predicted, f_predicted, expt_id)
                    refl_extend(rubbish, f_rubbish, expt_id)
                    f_experiments.append(experiment)
                    good_expt_count += 1
                else:
                    logger.info(
                        "Removing experiment %d: no reference reflections matched to predictions",
                        expt_id,
                    )

            reference = f_reference
            predicted = f_predicted
            experiments = f_experiments
            rubbish = f_rubbish

    # Select a random sample of the predicted reflections
    if not params.sampling.integrate_all_reflections:
        predicted = sample_predictions(experiments, predicted, params)

    # Compute the profile model - either load existing or compute
    # can raise RuntimeError
    experiments = ProfileModelFactory.create(params, experiments, reference)
    for expr in experiments:
        if expr.profile is None:
            raise ValueError("No profile information in experiment list")
    del reference

    # Compute the bounding box
    predicted.compute_bbox(experiments)

    # Create the integrator
    integrator = create_integrator(params, experiments, predicted)

    # Integrate the reflections
    reflections = integrator.integrate()

    # Remove unintegrated reflections
    if not params.output.output_unintegrated_reflections:
        keep = reflections.get_flags(reflections.flags.integrated, all=False)
        logger.info(
            "Removing %d unintegrated reflections of %d total",
            keep.count(False),
            keep.size(),
        )

        reflections = reflections.select(keep)

    # Append rubbish data onto the end
    if rubbish is not None and params.output.include_bad_reference:
        mask = flex.bool(len(rubbish), True)
        rubbish.unset_flags(mask, rubbish.flags.integrated_sum)
        rubbish.unset_flags(mask, rubbish.flags.integrated_prf)
        rubbish.set_flags(mask, rubbish.flags.bad_reference)
        reflections.extend(rubbish)

    # Correct integrated intensities for absorption correction, if necessary
    for abs_params in params.absorption_correction:
        if abs_params.apply and abs_params.algorithm == "fuller_kapton":
            from dials.algorithms.integration.kapton_correction import (
                multi_kapton_correction,
            )

            experiments, reflections = multi_kapton_correction(
                experiments, reflections, abs_params.fuller_kapton, logger=logger
            )()

    if params.significance_filter.enable:
        from dials.algorithms.integration.stills_significance_filter import (
            SignificanceFilter,
        )

        sig_filter = SignificanceFilter(params)
        filtered_refls = sig_filter(experiments, reflections)
        accepted_expts = ExperimentList()
        accepted_refls = flex.reflection_table()
        logger.info(
            "Removed %d reflections out of %d when applying significance filter",
            (reflections.size() - filtered_refls.size()),
            reflections.size(),
        )
        for expt_id, expt in enumerate(experiments):
            refls = filtered_refls.select(filtered_refls["id"] == expt_id)
            if refls:
                accepted_expts.append(expt)
                current_id = expt_id
                new_id = len(accepted_expts) - 1
                refls["id"] = flex.int(len(refls), new_id)
                if expt.identifier:
                    del refls.experiment_identifiers()[current_id]
                    refls.experiment_identifiers()[new_id] = expt.identifier
                accepted_refls.extend(refls)
            else:
                logger.info(
                    "Removed experiment %d which has no reflections left after applying significance filter",
                    expt_id,
                )

        if not accepted_refls:
            raise ValueError("No reflections left after applying significance filter")
        experiments = accepted_expts
        reflections = accepted_refls

    # Write a report if requested
    report = None
    if params.output.report is not None:
        report = integrator.report()

    return experiments, reflections, report


@show_mail_handle_errors()
def run(args=None, phil=working_phil):
    """Run the integration command line script."""
    usage = "usage: dials.integrate [options] models.expt"

    # Create the parser
    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        epilog=__doc__,
        read_experiments=True,
        read_reflections=True,
    )

    params, options = parser.parse_args(args=args, show_diff_phil=False)

    # Configure the logging
    dials.util.log.config(verbosity=options.verbose, logfile=params.output.log)

    logger.info(dials_version())

    # Log the PHIL diff
    # Log the diff phil
    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n%s", diff_phil)
    # Save phil parameters
    if params.output.phil is not None:
        with open(params.output.phil, "w") as outfile:
            outfile.write(parser.diff_phil.as_str())

    reference, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    if not reference and not experiments:
        parser.print_help()
        return
    if not experiments:
        sys.exit("No experiment list was specified")
    if not reference:
        reference = None
    elif len(reference) != 1:
        sys.exit("More than 1 reflection file was given")
    else:
        reference = reference[0]

    if reference and "shoebox" not in reference:
        sys.exit("Error: shoebox data missing from reflection table")

    try:
        experiments, reflections, report = run_integration(
            params, experiments, reference
        )
    except (ValueError, RuntimeError) as e:
        sys.exit(e)
    else:
        # Delete the shoeboxes used for intermediate calculations, if requested
        if params.integration.debug.delete_shoeboxes and "shoebox" in reflections:
            del reflections["shoebox"]

        logger.info(
            "Saving %d reflections to %s", reflections.size(), params.output.reflections
        )
        reflections.as_file(params.output.reflections)
        logger.info("Saving the experiments to %s", params.output.experiments)
        experiments.as_file(params.output.experiments)

        if report:
            report.as_file(params.output.report)


if __name__ == "__main__":
    run()



"""
This module defines a general cross validation function that can be used
with any valid CrossValidator. To use in a command line program, the
phil_scope should be included. Provided here is a description of the options,
using the example of running cross_validation in dials.scale

General description:
cross_validate runs the script defined by the CrossValidator, running each
option in turn, using a free set to score the model - the results are printed
in a table and the model with the lowest free set rmsd is indicated. For each
option, the analysis will be repeated nfolds times, with a different free set
chosen each time, and the final rmsds averaged. For full k-fold cross-validation,
nfolds should be set to 100/free_set_percentage, which would be nfolds=10 for
the default free_set_percentage=10.0.

Two different modes are currently supported, controlled by cross_validation_mode=;
1) cross_validation_mode=single
   dials.scale is run nfolds times for the user specified dials.scale options
2) cross_validation_mode=multi
   optimise a dials.scale parameter, specified by parameter= .The full phil path
   to the parameter must be specified, e.g. physical.decay_correction. The phil
   structure can be seen by running dials.scale -ce2
   parameter_values must also be specified as a string of space separated values,
   unless the dials.scale parameter type is bool.

Therefore one must choose:
  cross_validation_mode= (single or multi)
  parameter=             (a supported command line option of the script run by
                          CrossValidator, optional if cross_validation_mode=single)
  parameter_values=      (values to test, only optional if parameter= selects a
                          boolean command-line parameter)

For example
cross_validation_mode=multi parameter=physical.absorption_correction
cross_validation_mode=multi parameter=physical.decay_interval parameter_values="5.0 10.0 15.0"
cross_validation_mode=multi parameter=model parameter_values="array physical"
"""


from __future__ import annotations

import itertools
import logging
import time

from libtbx import phil

logger = logging.getLogger("dials")

phil_scope = phil.parse(
    """
  cross_validation {
    cross_validation_mode = multi single
      .type = choice
      .help = "Choose the cross validation running mode, for a full description"
              "see the module docstring. Choice is used for testing a parameter"
              "that can only have discreet values (a choice or bool phil parameter)."
              "Variable is used for testing a parameter that can have a float or"
              "int value (that is also not a 'choice' type). Single just performs"
              "cross validation on one parameter configuration."
      .expert_level = 2
    parameter = None
      .type = str
      .help = "Optimise a command-line parameter. The full phil path must be":
              "specified e.g. physical.absorption_correction. The option"
              "parameter_values must also be specified, unless the parameter is"
              "a True/False option."
      .expert_level = 2
    parameter_values = None
      .type = strings
      .help = "Parameter values to compare, entered as a string of space"
              "separated values."
      .expert_level = 2
    nfolds = 1
      .type = int(value_min=1)
      .help = "Number of cross-validation folds to perform. If nfolds > 1, the"
              "minimisation for each option is repeated nfolds times, with an"
              "incremental offset for the free set. The max number of folds"
              "allowed is 1/free_set_percentage; if set greater than this then"
              "the repetition will finish after 1/free_set_percentage folds."
      .expert_level = 2
  }
"""
)


def cross_validate(params, cross_validator):
    """Run cross validation script."""

    start_time = time.time()
    free_set_percentage = cross_validator.get_free_set_percentage(params)
    options_dict = {}

    if params.cross_validation.cross_validation_mode == "single":
        # just run the setup nfolds times
        cross_validator.create_results_dict(n_options=1)
        for n in range(params.cross_validation.nfolds):
            if n < 100.0 / free_set_percentage:
                params = cross_validator.set_free_set_offset(params, n)
                cross_validator.run_script(params, config_no=0)

    elif params.cross_validation.cross_validation_mode == "multi":
        # run each option nfolds times
        if params.cross_validation.parameter is None:
            raise ValueError(
                "parameter= must be set to specify what command line option should be optimised"
            )

        choice = params.cross_validation.parameter
        # #TODO extract allowed values to allow checking of user input

        # inspect the phil scope to see what the parameter type is e.g bool, int
        typ = cross_validator.get_parameter_type(choice)

        if typ == "bool" and not params.cross_validation.parameter_values:
            # values not specified, implied that should test both True and False
            options_dict[choice] = [True, False]
        else:
            if not params.cross_validation.parameter_values:
                raise ValueError(
                    "parameter_values= must be set to specify what options should be tested"
                )
            options_dict[choice] = []
            if typ == "bool":
                if (
                    "true" in params.cross_validation.parameter_values
                    or "True" in params.cross_validation.parameter_values
                ):
                    options_dict[choice].append(True)
                if (
                    "false" in params.cross_validation.parameter_values
                    or "False" in params.cross_validation.parameter_values
                ):
                    options_dict[choice].append(False)
            elif typ == "choice":
                for option in params.cross_validation.parameter_values:
                    options_dict[choice].append(option)
            elif typ == "int":
                for value in params.cross_validation.parameter_values:
                    options_dict[choice].append(int(value))
            elif typ == "float":
                for value in params.cross_validation.parameter_values:
                    options_dict[choice].append(float(value))
            else:
                raise ValueError("Error in interpreting parameter and parameter_values")

        # this code below should work for more than one parameter to be optimised,
        # but one cannot specify this yet from the command line
        keys, values = zip(*options_dict.items())

        cross_validator.create_results_dict(len(values[0]))
        cross_validator.set_results_dict_configuration(keys, values)

        for i, v in enumerate(itertools.product(*values)):
            e = dict(zip(keys, v))
            for k, val in e.items():
                params = cross_validator.set_parameter(params, k, val)
            for n in range(params.cross_validation.nfolds):
                if n < 100.0 / free_set_percentage:
                    params = cross_validator.set_free_set_offset(params, n)
                    cross_validator.run_script(params, config_no=i)

    else:
        raise ValueError("Error in interpreting mode and options.")

    st = cross_validator.interpret_results()
    logger.info("Summary of the cross validation analysis: \n %s", st.format())

    finish_time = time.time()
    logger.info(
        "\nCross-validation finished.\nTotal time taken: %.4fs ",
        finish_time - start_time,
    )
    logger.info("\n" + "=" * 80 + "\n")



""" Classes for scaling refinement engines.

Classes are inherited from the dials.refinement engine with a few
methods overwritten to use them with scaling code."""


from __future__ import annotations

import logging

from libtbx.phil import parse

from dials.algorithms.refinement.engine import (
    GaussNewtonIterations,
    LevenbergMarquardtIterations,
    SimpleLBFGS,
)
from dials.algorithms.scaling.scaling_utilities import log_memory_usage
from dials.util import tabulate

logger = logging.getLogger("dials")


TARGET_ACHIEVED = "RMSD target achieved"
RMSD_CONVERGED = "RMSD no longer decreasing"
STEP_TOO_SMALL = "Step too small"
OBJECTIVE_INCREASE = "Refinement failure: objective increased"
MAX_ITERATIONS = "Reached maximum number of iterations"
MAX_TRIAL_ITERATIONS = "Reached maximum number of consecutive unsuccessful trial steps"
DOF_TOO_LOW = "Not enough degrees of freedom to refine"

scaling_refinery_phil_str = """
scaling_refinery
  .help = "Parameters to configure the refinery"
  .expert_level = 1
{
  engine = *SimpleLBFGS GaussNewton LevMar
    .help = "The minimisation engine to use for the main scaling algorithm"
    .type = choice
  refinement_order = *concurrent consecutive
    .type = choice
    .help = "Choice of whether to refine all model components concurrently, or"
             "in a consecutive order as allowed/defined by the scaling model."
    .expert_level = 2
  max_iterations = None
    .help = "Maximum number of iterations in refinement before termination."
            "None implies the engine supplies its own default."
    .type = int(value_min=1)
  rmsd_tolerance = 0.0001
    .type = float(value_min=1e-6)
    .help = "Tolerance at which to stop scaling refinement. This is a relative
            value, the convergence criterion is (rmsd[i] - rmsd[i-1])/rmsd[i] <
            rmsd_tolerance."
  full_matrix_engine = GaussNewton *LevMar
    .help = "The minimisation engine to use for a full matrix round of
             minimisation after the main scaling, in order to determine
             error estimates."
    .type = choice
  full_matrix_max_iterations = None
    .help = "Maximum number of iterations before termination in the full matrix
             minimisation round."
            "None implies the engine supplies its own default."
    .type = int(value_min=1)
}
"""
"""refinery_log = None
    .help = "Filename for an optional log that a minimisation engine may use"
            "to write additional information"
    .type = path

  journal
    .help = "Extra items to track in the refinement history"
  {
    track_step = False
      .help = "Record parameter shifts history in the refinement journal, if"
              "the engine supports it."
      .type = bool

    track_gradient = False
      .help = "Record parameter gradients history in the refinement journal, if"
              "the engine supports it."
      .type = bool

    track_parameter_correlation = False
      .help = "Record correlation matrix between columns of the Jacobian for"
              "each step of refinement."
      .type = bool

    track_condition_number = False
      .help = "Record condition number of the Jacobian for each step of "
              "refinement."
      .type = bool

    track_out_of_sample_rmsd = False
      .type = bool
      .help = "Record RMSDs calculated using the refined experiments with"
              "reflections not used in refinement at each step. Only valid if a"
              "subset of input reflections was taken for refinement"
  }
}
"""
scaling_refinery_phil_scope = parse(scaling_refinery_phil_str)


def scaling_refinery(
    engine, scaler, target, prediction_parameterisation, max_iterations
):
    """Return the correct engine based on phil parameters."""
    if engine == "SimpleLBFGS":
        return ScalingSimpleLBFGS(
            scaler,
            target=target,
            prediction_parameterisation=prediction_parameterisation,
            max_iterations=max_iterations,
        )
    elif engine == "GaussNewton":
        return ScalingGaussNewtonIterations(
            scaler,
            target=target,
            prediction_parameterisation=prediction_parameterisation,
            max_iterations=max_iterations,
        )
    elif engine == "LevMar":
        return ScalingLevenbergMarquardtIterations(
            scaler,
            target=target,
            prediction_parameterisation=prediction_parameterisation,
            max_iterations=max_iterations,
        )


def print_step_table(refinery):
    """print useful output about refinement steps in the form of a simple table"""

    logger.info("\nRefinement steps:")

    header = ["Step", "Nref"]
    for (name, units) in zip(refinery._target.rmsd_names, refinery._target.rmsd_units):
        header.append(name + "\n(" + units + ")")

    rows = []
    for i in range(refinery.history.get_nrows()):
        rmsds = list(refinery.history["rmsd"][i])
        rows.append(
            [str(i), str(refinery.history["num_reflections"][i])]
            + [f"{r:.5g}" for r in rmsds]
        )

    logger.info(tabulate(rows, header))
    logger.info(refinery.history.reason_for_termination)


class ScalingRefinery:
    "mixin class to add extra return method"

    def __init__(self, scaler, target, prediction_parameterisation, *args, **kwargs):
        self._target = target
        self._scaler = scaler
        self._rmsd_tolerance = scaler.params.scaling_refinery.rmsd_tolerance
        self._parameters = prediction_parameterisation

    def print_step_table(self):
        print_step_table(self)

    @property
    def rmsd_tolerance(self):
        return self._rmsd_tolerance

    def set_tolerance(self, tolerance):
        """Set a tolerance."""
        self._rmsd_tolerance = tolerance

    def test_rmsd_convergence(self):
        """Test for convergence of RMSDs"""

        # http://en.wikipedia.org/wiki/
        # Non-linear_least_squares#Convergence_criteria
        try:
            r1 = self.history["rmsd"][-1]
            r2 = self.history["rmsd"][-2]
        except IndexError:
            return False

        if r2[0] > 0:
            return abs((r2[0] - r1[0]) / r2[0]) < self._rmsd_tolerance
        else:
            return True

    def prepare_for_step(self):
        """Update the parameterisation and prepare the target function. Overwrites
        the prepare_for_step method from refinery to direct the updating away from
        the target function to the update_for_minimisation method."""

        x = self.x

        # set current parameter values
        self._parameters.set_param_vals(x)

        return

    def update_journal(self):
        """Append latest step information to the journal attributes"""

        # add step quantities to journal
        self.history.add_row()
        self.history.set_last_cell("num_reflections", self._scaler.Ih_table.size)
        self.history.set_last_cell(
            "rmsd", self._target.rmsds(self._scaler.Ih_table, self._parameters)
        )
        self.history.set_last_cell(
            "parameter_vector", self._parameters.get_param_vals()
        )
        self.history.set_last_cell("objective", self._f)
        if "gradient" in self.history:
            self.history.set_last_cell("gradient", self._g)
        return


class ScalingSimpleLBFGS(ScalingRefinery, SimpleLBFGS):
    """Adapt Refinery for L-BFGS minimiser"""

    def __init__(self, scaler, *args, **kwargs):
        logger.info("Performing a round of scaling with an LBFGS minimizer. \n")
        ScalingRefinery.__init__(self, scaler, *args, **kwargs)
        SimpleLBFGS.__init__(self, *args, **kwargs)

    def compute_functional_gradients_and_curvatures(self):
        """overwrite method to avoid calls to 'blocks' methods of target"""
        self.prepare_for_step()

        work_blocks = self._scaler.get_blocks_for_minimisation()

        f = []
        gi = []
        for block_id, block in enumerate(work_blocks):
            self._scaler.update_for_minimisation(self._parameters, block_id)
            fb, gb = self._parameters.compute_functional_gradients(block)
            f.append(fb)
            gi.append(gb)
        """task_results = easy_mp.parallel_map(
            func=self._target.compute_functional_gradients,
            iterable=blocks,
            processes=self._scaler.params.scaling_options.nproc,
            method="multiprocessing",
            preserve_exception_message=True
        )
        f, gi = zip(*task_results)"""

        f = sum(f)
        g = gi[0]
        for i in range(1, len(gi)):
            g += gi[i]

        restraints = self._parameters.compute_restraints_functional_gradients(
            self._parameters
        )

        if restraints:
            f += restraints[0]
            g += restraints[1]
        logger.debug("Functional : %s", f)
        logger.debug("Gradients : %s", list(g))
        log_memory_usage()
        logger.debug("\n")
        return f, g, None


class ScalingLstbxBuildUpMixin(ScalingRefinery):
    """Mixin class to overwrite the build_up method in AdaptLstbx"""

    def build_up(self, objective_only=False):
        "overwrite method from Adaptlstbx"
        # set current parameter values
        self.prepare_for_step()

        # Reset the state to construction time, i.e. no equations accumulated
        self.reset()

        work_blocks = self._scaler.get_blocks_for_minimisation()

        # observation terms
        if objective_only:
            for block_id, block in enumerate(work_blocks):
                self._scaler.update_for_minimisation(self._parameters, block_id)
                residuals, weights = self._parameters.compute_residuals(block)
                self.add_residuals(residuals, weights)
        else:
            self._jacobian = None

            for block_id, block in enumerate(work_blocks):
                self._scaler.update_for_minimisation(self._parameters, block_id)
                (
                    residuals,
                    jacobian,
                    weights,
                ) = self._parameters.compute_residuals_and_gradients(block)
                self.add_equations(residuals, jacobian, weights)
            """task_results = easy_mp.pool_map(
                fixed_func=self._target.compute_residuals_and_gradients,
                iterable=blocks,
                processes=self._scaler.params.scaling_options.nproc
            )
            for result in task_results:
                self.add_equations(result[0], result[1], result[2])"""

        restraints = self._parameters.compute_restraints_residuals_and_gradients(
            self._parameters
        )
        if restraints:
            if objective_only:
                self.add_residuals(restraints[0], restraints[2])
            else:
                self.add_equations(restraints[0], restraints[1], restraints[2])
        logger.debug("added equations for all blocks")
        log_memory_usage()
        logger.debug("\n")
        return


class ScalingGaussNewtonIterations(ScalingLstbxBuildUpMixin, GaussNewtonIterations):
    """Refinery implementation, using lstbx Gauss Newton iterations"""

    # defaults that may be overridden
    gradient_threshold = 1.0e-10
    step_threshold = None
    damping_value = 0.0007
    max_shift_over_esd = 15
    convergence_as_shift_over_esd = 1e-5

    def __init__(
        self,
        scaler,
        target,
        prediction_parameterisation,
        constraints_manager=None,
        log=None,
        tracking=None,
        max_iterations=20,
    ):
        logger.info("Performing a round of scaling with a Gauss-Newton minimizer.\n")
        ScalingLstbxBuildUpMixin.__init__(
            self, scaler, target, prediction_parameterisation
        )
        GaussNewtonIterations.__init__(
            self,
            target,
            prediction_parameterisation,
            constraints_manager,
            log=log,
            tracking=tracking,
            max_iterations=max_iterations,
        )


class ScalingLevenbergMarquardtIterations(
    ScalingLstbxBuildUpMixin, LevenbergMarquardtIterations
):
    """Refinery implementation, employing lstbx Levenberg Marquadt
    iterations"""

    def __init__(
        self,
        scaler,
        target,
        prediction_parameterisation,
        constraints_manager=None,
        log=None,
        tracking=None,
        max_iterations=20,
    ):
        logger.info(
            "Performing a round of scaling with a Levenberg-Marquardt minimizer.\n"
        )
        ScalingLstbxBuildUpMixin.__init__(
            self, scaler, target, prediction_parameterisation
        )
        LevenbergMarquardtIterations.__init__(
            self,
            target,
            prediction_parameterisation,
            constraints_manager,
            log=log,
            tracking=tracking,
            max_iterations=max_iterations,
        )



from __future__ import annotations

import contextlib
import dataclasses
import faulthandler
import functools
import io
import os
import pathlib
import signal
import sys
from typing import Optional

import tabulate as _tabulate
import tqdm

from libtbx.utils import Sorry

# Coloured exit code support on windows
try:
    import colorama
except ImportError:
    pass
else:
    colorama.init()

__all__ = [
    "debug_console",
    "debug_context_manager",
    "progress",
    "show_mail_handle_errors",
    "Sorry",
    "tabulate",
]

# Define the default tablefmt in dials
tabulate = functools.partial(_tabulate.tabulate, tablefmt="psql")
functools.update_wrapper(tabulate, _tabulate.tabulate)

# Customisable progressbar decorator for iterators.
#
# Utilizes the progress bar from the tqdm package, with modified defaults:
#   - By default, resize when terminal is resized (dynamic-columns)
#   - By default, disables the progress bar for non-tty output
#   - By default, the progress bar will be removed after completion
#
# Usage:
#   >>> from dials.util import progress
#   >>> for i in progress(range(10)):
#   ...     ...
#
# See https://github.com/tqdm/tqdm for more in-depth usage and options.
progress = functools.partial(tqdm.tqdm, disable=None, dynamic_ncols=True, leave=False)
functools.update_wrapper(
    functools.partial(tqdm.tqdm, disable=None, dynamic_ncols=True, leave=False),
    tqdm.tqdm,
)


def debug_console():
    """Start python console at the current code point."""

    # use exception trick to pick up the current frame
    try:
        raise RuntimeError()
    except RuntimeError:
        frame = sys.exc_info()[2].tb_frame.f_back

    # evaluate commands in current namespace
    namespace = frame.f_globals.copy()
    namespace.update(frame.f_locals)

    try:
        # Start IPython console if IPython is available.
        from IPython import embed

        embed(user_ns=namespace, colors="neutral")
    except ImportError:
        # Otherwise use basic python console
        import code

        code.interact(banner="=" * 80, local=namespace)


def debug_context_manager(original_context_manager, name="", log_func=None):
    """A wrapper to help debugging arbitrary context managers. For example
    instead of
      lock = threading.RLock():
    use
      lock = debug_context_manager(threading.RLock())
    and all calls
      with lock:
        ...
    will produce (hopefully) helpful debug output.

    :param original_context_manager: Some context manager to be wrapped.
    :param name: A name for the context manager, that will be printed in the
                 debug output.
    :param log_func: optional log function. If not specified debug output
                     will be printed to sys.stderr.
    :return: A context manager wrapping the original context manager.
    """
    if not log_func:

        def log_func(output):
            sys.stderr.write(output)
            sys.stderr.flush()

    import multiprocessing
    import threading
    from datetime import datetime

    class DCM:
        def __init__(self, name, log_func):
            self._ocm = original_context_manager
            if name:
                self._name = f"{name}-{hash(original_context_manager)}"
            else:
                self._name = str(hash(original_context_manager))
            self._log_func = log_func

        def format_event(self, event):
            return f"{datetime.now().strftime('%H:%M:%S.%f')} {self._name}: {event}\n"

        def log(self, event):
            self._log_func(self.format_event(event))

        def __enter__(self, *args, **kwargs):
            try:
                raise Exception()
            except Exception:
                parent = sys.exc_info()[2].tb_frame.f_back
            call_code = f"{parent.f_code.co_filename}:{parent.f_lineno}"
            call_process = multiprocessing.current_process().name
            call_thread = threading.currentThread().getName()
            self.log(f"Knock {call_process}:{call_thread} {call_code}")
            z = self._ocm.__enter__(*args, **kwargs)
            self.log(f"Enter {call_process}:{call_thread} {call_code}")
            return z

        def __exit__(self, *args, **kwargs):
            call_process = multiprocessing.current_process().name
            call_thread = threading.currentThread().getName()
            self.log(f"Exit {call_process}:{call_thread}")
            self._ocm.__exit__(*args, **kwargs)
            self.log(f"Left {call_process}:{call_thread}")

    return DCM(name, log_func)


@contextlib.contextmanager
def enable_faulthandler():
    # Don't clobber someone elses faulthandler settings
    if not faulthandler.is_enabled():
        # SIGUSR2 not available on windows
        # The attached STDERR might not support what faulthandler wants
        with contextlib.suppress(AttributeError, io.UnsupportedOperation):
            faulthandler.enable()
            faulthandler.register(signal.SIGUSR2, all_threads=True)
    yield


@contextlib.contextmanager
def show_mail_on_error():
    try:
        yield
    except Exception as e:
        text = "Please report this error to dials-support@lists.sourceforge.net:"
        if issubclass(e.__class__, Sorry):
            raise

        if len(e.args) == 1:
            e.args = (f"{text} {e.args[0]}",)
        else:
            e.args = (text, *e.args)

        raise


@contextlib.contextmanager
def make_sys_exit_red():
    """Make sys.exit call messages red, if appropriate"""
    try:
        yield
    except SystemExit as e:
        # Don't colour if requested not to, or user formatted already
        if (
            isinstance(e.code, str)
            and "NO_COLOR" not in os.environ
            and sys.stderr.isatty()
            and "\033" not in e.code
        ):
            e.code = f"\033[31m{e.code}\033[0m"
        raise


@contextlib.contextmanager
def show_mail_handle_errors():
    """
    Handle showing errors to the user, including contact email, and colors.

    If a stack trace occurs, the user will be give the contact email to report
    the error to. If a sys.exit is caught, it will be converted to red, if
    appropriate.

    On Windows this additionally changes the stdout/stderr encoding to UTF8,
    which will only have an effect when the program output is redirected to
    a file or another process. This change will outlive the calling context.
    """
    if os.name == "nt":
        import dxtbx.util

        dxtbx.util.encode_output_as_utf8()
    with enable_faulthandler(), make_sys_exit_red(), show_mail_on_error():
        yield


@dataclasses.dataclass
class HTCondorJobClassAd:
    cpus_provisioned: Optional[int] = None
    memory_provisioned: Optional[float] = None


def parse_htcondor_job_classad(filepath: pathlib.Path) -> HTCondorJobClassAd:
    with filepath.open() as fh:
        memory_provisioned = None
        cpus_provisioned = None
        for line in fh.readlines():
            line = line.lower().strip()
            if line.startswith("memoryprovisioned") and "=" in line:
                memory_provisioned = float(line.split("=")[1])
            elif line.startswith("cpusprovisioned") and "=" in line:
                cpus_provisioned = int(line.split("=")[1])
        return HTCondorJobClassAd(
            cpus_provisioned=cpus_provisioned,
            memory_provisioned=memory_provisioned,
        )



from __future__ import annotations

import procrunner
import pytest

from dials.array_family import flex


@pytest.fixture(scope="session")
def reflections(tmp_path_factory):
    # Make a dummy reflection table for the test setting some values and flags
    rt = flex.reflection_table.empty_standard(6)
    rt["iobs"] = flex.size_t_range(len(rt))
    rt["panel"] = flex.size_t_range(len(rt))
    rt["id"] = flex.int([0] * 5 + [1])
    rt["d"] = flex.double([50, 40, 3.0, 2.5, 2.0, 1.0])
    mask1 = flex.bool([True] * 3 + [False] * 3)
    mask2 = flex.bool([True, False] * 3)
    rt.set_flags(mask1, rt.flags.integrated)
    rt.set_flags(mask2, rt.flags.reference_spot)
    tmp_path = tmp_path_factory.mktemp("filter_reflections")
    rt_name = tmp_path / "test_refs.refl"
    rt.as_file(rt_name)
    return rt_name


def test_filter_reflections_flag_expression(reflections, tmp_path):
    result = procrunner.run(
        [
            "dials.filter_reflections",
            reflections,
            "flag_expression='integrated & ~reference_spot'",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    ref = flex.reflection_table.from_file(tmp_path / "filtered.refl")
    # The test selects only the 2nd reflection
    assert len(ref) == 1
    assert list(ref["iobs"]) == [1]


def test_filter_reflections_by_experiment_id(reflections, tmp_path):
    result = procrunner.run(
        ["dials.filter_reflections", reflections, "id=0"], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr
    ref = flex.reflection_table.from_file(tmp_path / "filtered.refl")
    # The test selects only the first five reflections
    assert len(ref) == 5
    assert list(ref["iobs"]) == [0, 1, 2, 3, 4]


def test_filter_reflections_by_panel(reflections, tmp_path):
    result = procrunner.run(
        ["dials.filter_reflections", reflections, "panel=5"], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr
    ref = flex.reflection_table.from_file(tmp_path / "filtered.refl")
    # The test selects only the last reflection
    assert len(ref) == 1
    assert list(ref["iobs"]) == [5]


def test_filter_reflections_by_resolution(reflections, tmp_path):
    result = procrunner.run(
        ["dials.filter_reflections", reflections, "d_max=3.0", "d_min=2.0"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    ref = flex.reflection_table.from_file(tmp_path / "filtered.refl")
    # The test selects only the 3rd, 4th and 5th reflections
    assert len(ref) == 3
    assert list(ref["iobs"]) == [2, 3, 4]


def test_filter_reflections_printing_analysis(reflections, tmp_path):
    result = procrunner.run(
        ["dials.filter_reflections", reflections], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr


@pytest.mark.parametrize(
    "experiment,reflections,args,expected",
    [
        (
            "imported_experiments.json",
            "strong.pickle",
            ["ice_rings.filter=True"],
            615,
        ),
        (
            "imported_experiments.json",
            "strong.pickle",
            ["d_min=3.1", "d_max=20"],
            153,
        ),
        (
            "experiments.json",
            "integrated.refl",
            ["d_min=2", "d_max=20"],
            371,
        ),
    ],
)
def test_filter_reflections(
    experiment, reflections, args, expected, dials_data, tmp_path
):
    dataset = dials_data("centroid_test_data", pathlib=True)
    result = procrunner.run(
        [
            "dials.filter_reflections",
            dataset / experiment,
            dataset / reflections,
        ]
        + args,
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    filtered_refl = flex.reflection_table.from_file(tmp_path / "filtered.refl")
    assert len(filtered_refl) == expected


def test_filter_reflections_by_reflection_index(reflections, tmp_path):
    result = procrunner.run(
        ["dials.filter_reflections", reflections, "remove_by_index=0,3,5"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    ref = flex.reflection_table.from_file(tmp_path / "filtered.refl")
    # The test removed the reflections 0, 3 and 5
    assert len(ref) == 3
    assert list(ref["iobs"]) == [1, 2, 4]



# LIBTBX_SET_DISPATCHER_NAME libtbx.precommit

from __future__ import annotations

import dials.precommitbx.installer
import dials.util


@dials.util.show_mail_handle_errors()
def run(_=None):
    dials.precommitbx.installer.main()


if __name__ == "__main__":
    run()



from __future__ import annotations

import os

import pytest

from dials.algorithms.refinement.outlier_detection import CentroidOutlierFactory
from dials.algorithms.refinement.outlier_detection.outlier_base import phil_scope
from dials.array_family import flex


@pytest.mark.parametrize(
    "method,colnames,expected_nout",
    [
        ("tukey", ("x_resid", "y_resid", "phi_resid"), 34),
        pytest.param("mcd", ("x_resid", "y_resid", "phi_resid"), 35),
        pytest.param("mcd", ("x_resid", "y_resid", "phi_resid"), 35),
        pytest.param(
            "sauter_poon", ("miller_index", "xyzobs.px.value", "xyzcal.px"), 34
        ),
    ],
)
def test_centroid_outlier(dials_regression, method, colnames, expected_nout):

    flex.set_random_seed(42)
    data_dir = os.path.join(
        dials_regression, "refinement_test_data", "centroid_outlier"
    )
    residuals = flex.reflection_table.from_file(
        os.path.join(data_dir, "residuals.refl")
    )
    params = phil_scope.extract()
    params.outlier.algorithm = method
    params.outlier.sauter_poon.px_sz = (0.1, 0.1)  # must be set for SauterPoon
    outlier_detector = CentroidOutlierFactory.from_parameters_and_colnames(
        params, colnames
    )
    outlier_detector(residuals)
    outliers = residuals.get_flags(residuals.flags.centroid_outlier)

    assert outliers.count(True) == expected_nout



# LIBTBX_PRE_DISPATCHER_INCLUDE_SH export PHENIX_GUI_ENVIRONMENT=1


from __future__ import annotations

import collections
import copy
import json
import sys

import iotbx.phil
from cctbx import sgtbx
from dxtbx.model import MultiAxisGoniometer
from rstbx.cftbx.coordinate_frame_helpers import align_reference_frame
from scitbx import matrix

import dials.util
from dials.algorithms.refinement import rotation_decomposition
from dials.util import tabulate
from dials.util.options import ArgumentParser, flatten_experiments

help_message = """
Calculation of possible goniometer settings for re-alignment of crystal axes.

By default the program will attempt to calculate possible goniometer settings to
align the primary crystal axes with the principle goniometer axis. Optionally
vectors to align may be specified in pairs using the vector= parameter. If
mode=main (default), then the first vector of a pair will be aligned along the
principle goniometer axis, with the second vector placed in the plane containing
the beam vector and the principle goniometer axis. This allows for rotation of
the crystal around a principle crystal axis. If mode=cusp, then the first vector
is aligned perpendicular to the beam and the principle goniometer axis, with the
second vector placed in the plane containing the first crystal vector and the
principle goniometer axis.

Examples::

  dials.align_crystal models.expt

  dials.align_crystal models.expt vector=0,0,1 vector=0,1,0

  dials.align_crystal models.expt frame=direct
"""

phil_scope = iotbx.phil.parse(
    """
space_group = None
  .type = space_group
align {
  mode = *main cusp
    .type = choice
  crystal {
    vector = None
      .type = str
      .multiple = True
    frame = *reciprocal direct
      .type = choice
  }
}
output {
  json = align_crystal.json
    .type = path
}
"""
)


a_star = matrix.col((1.0, 0.0, 0.0))
b_star = matrix.col((0.0, 1.0, 0.0))
c_star = matrix.col((0.0, 0.0, 1.0))

a = matrix.col((1.0, 0.0, 0.0))
b = matrix.col((0.0, 1.0, 0.0))
c = matrix.col((0.0, 0.0, 1.0))


def smallest_angle(angle):
    return min(abs(angle), abs(180 - angle))


def describe(vector, space_group, reciprocal=True):

    vector_names = {a.elems: "a", b.elems: "b", c.elems: "c"}

    v = vector.elems
    if v in vector_names:
        vstr = vector_names[v]
        if reciprocal:
            vstr += "*"
            t = axis_type(vector, space_group)
            if t is not None:
                vstr = "%s (%i-fold)" % (vstr, t)
    else:
        vstr = str(v)
    return vstr


def axis_type(vector, space_group):
    v = vector.elems

    axis_t = None
    for op in space_group.smx():
        info = op.r().info()
        ev = info.ev()
        if v == ev:
            if axis_t is None or info.type() > axis_t:
                axis_t = info.type()

    return axis_t


class align_crystal:

    vector_names = {a.elems: "a", b.elems: "b", c.elems: "c"}

    def __init__(self, experiment, vectors, frame="reciprocal", mode="main"):
        self.experiment = experiment
        self.vectors = vectors
        self.frame = frame
        self.mode = mode

        gonio = experiment.goniometer

        self.s0 = matrix.col(self.experiment.beam.get_s0())
        self.rotation_axis = matrix.col(gonio.get_rotation_axis())

        if not isinstance(gonio, MultiAxisGoniometer):
            raise ValueError("Only MultiAxisGoniometer models supported")
        axes = gonio.get_axes()
        if len(axes) != 3:
            raise ValueError("Only 3-axis goniometers supported")
        e1, e2, e3 = (matrix.col(e) for e in reversed(axes))

        # fixed_rotation = matrix.sqr(gonio.get_fixed_rotation())
        # setting_rotation = matrix.sqr(gonio.get_setting_rotation())
        # rotation_axis = matrix.col(gonio.get_rotation_axis_datum())
        # rotation_matrix = rotation_axis.axis_and_angle_as_r3_rotation_matrix(
        #    experiment.scan.get_oscillation()[0], deg=True
        # )

        results = []

        # from https://github.com/legrandp/xdsme/blob/master/XOalign/XOalign.py#L427
        #  referential_permutations sign permutations for four permutations of
        #        parallel/antiparallel (rotation axis & beam)
        #    y1 // e1, y2 // beamVector;  y1 anti// e1, y2 // beamVector
        #    y1 // e1, y2 anti// beamVector;  y1 anti// e1, y2 anti// beamVector

        ex = matrix.col((1, 0, 0))
        ey = matrix.col((0, 1, 0))
        ez = matrix.col((0, 0, 1))

        referential_permutations = (
            [ex, ey, ez],
            [-ex, -ey, ez],
            [ex, -ey, -ez],
            [-ex, ey, -ez],
        )

        for (v1_, v2_) in self.vectors:
            result_dictionary = collections.defaultdict(list)
            results.append((v1_, v2_, result_dictionary))
            space_group = self.experiment.crystal.get_space_group()
            for smx in list(space_group.smx())[:]:
                crystal = copy.deepcopy(self.experiment.crystal)
                cb_op = sgtbx.change_of_basis_op(smx)
                crystal = crystal.change_basis(cb_op)

                # Goniometer datum setting [D] at which the orientation was determined
                # D = (setting_rotation * rotation_matrix * fixed_rotation).inverse()

                # The setting matrix [U] will vary with the datum setting according to
                # [U] = [D] [U0]
                U = matrix.sqr(crystal.get_U())

                # XXX In DIALS recorded U is equivalent to U0 - D is applied to U inside
                # prediction
                U0 = U

                B = matrix.sqr(crystal.get_B())

                if self.frame == "direct":
                    B = B.inverse().transpose()

                v1_0 = U0 * B * v1_
                v2_0 = U0 * B * v2_

                # c  (b) The laboratory frame vectors l1 & l2 are normally specified with the
                # c MODE command: MODE MAIN (the default) sets l1 (along which v1 will be
                # c placed) along the principle goniostat axis e1 (Omega), and l2 along
                # c the beam s0. This allows rotation for instance around a principle axis.
                # c The other mode is MODE CUSP, which puts l1 (v1) perpendicular to the
                # c beam (s0) and the e1 (Omega) axis, and l2 (v2) in the plane containing
                # c l1 & e1 (ie l1 = e1 x s0, l2 = e1).

                if self.mode == "cusp":
                    l1 = self.rotation_axis.cross(self.s0)
                    l2 = self.rotation_axis
                else:
                    l1 = self.rotation_axis.normalize()
                    l3 = l1.cross(self.s0).normalize()
                    l2 = l1.cross(l3)

                for perm in referential_permutations:
                    S = matrix.sqr(perm[0].elems + perm[1].elems + perm[2].elems)
                    R = align_reference_frame(v1_0, S * l1, v2_0, S * l2)
                    solutions = (
                        rotation_decomposition.solve_r3_rotation_for_angles_given_axes(
                            R, e1, e2, e3, return_both_solutions=True, deg=True
                        )
                    )

                    if solutions is None:
                        continue

                    result_dictionary[smx].extend(solutions)

        self.all_solutions = results

        self.unique_solutions = collections.defaultdict(list)
        for v1, v2, result in results:
            for solutions in result.values():
                for solution in solutions:
                    k = tuple(round(a, 3) for a in solution[1:])
                    self.unique_solutions.setdefault(k, [])
                    if all(v1 != z1 or v2 != z2 for z1, z2 in self.unique_solutions[k]):
                        self.unique_solutions[k].append((v1, v2))

    def _vector_as_str(self, v):
        v = v.elems
        if v in self.vector_names:
            vstr = self.vector_names[v]
            if self.frame == "reciprocal":
                vstr += "*"
        else:
            vstr = str(v)
        return vstr

    def as_json(self, filename=None):
        names = self.experiment.goniometer.get_names()
        space_group = self.experiment.crystal.get_space_group()
        solutions = [
            {
                "primary_axis": [self._vector_as_str(v1) for v1, v2 in solns],
                "secondary_axis": [self._vector_as_str(v2) for v1, v2 in solns],
                "primary_axis_type": [axis_type(v1, space_group) for v1, v2 in solns],
                "secondary_axis_type": [axis_type(v2, space_group) for v1, v2 in solns],
                names[1]: angles[0],
                names[0]: angles[1],
            }
            for angles, solns in self.unique_solutions.items()
        ]
        d = {"solutions": solutions, "goniometer": self.experiment.goniometer.to_dict()}

        if filename:
            with open(filename, "w") as fh:
                json.dump(d, fh, indent=2)
        else:
            return json.dumps(d, indent=2)

    def __str__(self):
        U = matrix.sqr(self.experiment.crystal.get_U())
        B = matrix.sqr(self.experiment.crystal.get_B())

        a_star_ = U * B * a_star
        b_star_ = U * B * b_star
        c_star_ = U * B * c_star

        Binvt = B.inverse().transpose()

        a_ = U * Binvt * a
        b_ = U * Binvt * b
        c_ = U * Binvt * c

        names = self.experiment.goniometer.get_names()
        axes = self.experiment.goniometer.get_axes()
        rows = [["Experimental axis", "a*", "b*", "c*"]]
        rows.append(
            [names[0]]
            + [
                f"{smallest_angle(axis.angle(matrix.col(axes[0]), deg=True)):.3f}"
                for axis in (a_star_, b_star_, c_star_)
            ]
        )
        rows.append(
            ["Beam"]
            + [
                f"{smallest_angle(axis.angle(self.s0, deg=True)):.3f}"
                for axis in (a_star_, b_star_, c_star_)
            ]
        )
        rows.append(
            [names[2]]
            + [
                f"{smallest_angle(axis.angle(matrix.col(axes[2]), deg=True)):.3f}"
                for axis in (a_star_, b_star_, c_star_)
            ]
        )

        output = []
        output.append(
            "Angles between reciprocal cell axes and principal experimental axes:"
        )
        output.append(tabulate(rows, headers="firstrow"))
        output.append("")

        rows = [["Experimental axis", "a", "b", "c"]]
        rows.append(
            [names[0]]
            + [
                f"{smallest_angle(axis.angle(matrix.col(axes[0]), deg=True)):.3f}"
                for axis in (a_, b_, c_)
            ]
        )
        rows.append(
            ["Beam"]
            + [
                f"{smallest_angle(axis.angle(self.s0, deg=True)):.3f}"
                for axis in (a_, b_, c_)
            ]
        )
        rows.append(
            [names[2]]
            + [
                f"{smallest_angle(axis.angle(matrix.col(axes[2]), deg=True)):.3f}"
                for axis in (a_, b_, c_)
            ]
        )
        output.append("Angles between unit cell axes and principal experimental axes:")
        output.append(tabulate(rows, headers="firstrow"))
        output.append("")

        names = self.experiment.goniometer.get_names()

        space_group = self.experiment.crystal.get_space_group()
        reciprocal = self.frame == "reciprocal"
        rows = []
        for angles, vector_pairs in self.unique_solutions.items():
            v1, v2 = list(vector_pairs)[0]
            rows.append(
                (
                    describe(v1, space_group, reciprocal=reciprocal),
                    describe(v2, space_group, reciprocal=reciprocal),
                    f"{angles[0]: 7.3f}",
                    f"{angles[1]: 7.3f}",
                )
            )
        rows = [("Primary axis", "Secondary axis", names[1], names[0])] + sorted(rows)
        output.append("Independent solutions:")
        output.append(tabulate(rows, headers="firstrow"))

        return "\n".join(output)


@dials.util.show_mail_handle_errors()
def run(args=None):
    usage = "dials.align_crystal [options] models.expt"
    parser = ArgumentParser(
        usage=usage,
        phil=phil_scope,
        read_experiments=True,
        check_format=False,
        epilog=help_message,
    )

    params, options = parser.parse_args(args, show_diff_phil=True)
    experiments = flatten_experiments(params.input.experiments)

    if len(experiments) == 0:
        parser.print_help()
        exit(0)

    expt = experiments[0]

    if params.space_group is not None:
        expt.crystal.set_space_group(params.space_group.group())

    if len(params.align.crystal.vector):
        frame = None
        assert len(params.align.crystal.vector) % 2 == 0
        vectors = []

        name_to_vectors = {
            "a": (a, "direct"),
            "b": (b, "direct"),
            "c": (c, "direct"),
            "a*": (a_star, "reciprocal"),
            "b*": (b_star, "reciprocal"),
            "c*": (c_star, "reciprocal"),
        }

        for v in params.align.crystal.vector:
            v = v.strip()
            if v in name_to_vectors:
                v, frame_ = name_to_vectors[v]
                assert frame is None or frame == frame_
                frame = frame_
            else:
                v = v.replace(",", " ").strip().split()
                assert len(v) == 3
                v = matrix.col([float(v_) for v_ in v])
                if frame is None:
                    frame = params.align.crystal.frame

            vectors.append(v)
        vectors = [
            (vectors[2 * i], vectors[2 * i + 1]) for i in range(len(vectors) // 2)
        ]
    elif params.align.crystal.frame == "direct":
        frame = params.align.crystal.frame
        vectors = ((a, b), (a, c), (b, a), (b, c), (c, a), (c, b))

    else:
        frame = "reciprocal"
        vectors = (
            (a_star, b_star),  # a*, b*
            (a_star, c_star),  # a*, c*
            (b_star, a_star),  # b*, a*
            (b_star, c_star),  # b*, c*
            (c_star, a_star),  # c*, a*
            (c_star, b_star),  # c*, b*
        )

    try:
        result = align_crystal(expt, vectors, frame=frame, mode=params.align.mode)
        print(result)
    except ValueError as e:
        sys.exit(e)
    if params.output.json is not None:
        result.as_json(filename=params.output.json)


if __name__ == "__main__":
    run()
